@@ -34,11 +34,10 @@ For paired-end reads:
     cutadapt -a ADAPT1 -A ADAPT2 [options] -o out1.fastq -p out2.fastq in1.fastq in2.fastq
 
 Replace "ADAPTER" with the actual sequence of your 3' adapter. IUPAC wildcard
-characters are supported. The reverse complement is *not* automatically
-searched. All reads from input.fastq will be written to output.fastq with the
-adapter sequence removed. Adapter matching is error-tolerant. Multiple adapter
-sequences can be given (use further -a options), but only the best-matching
-adapter will be removed.
+characters are supported. All reads from input.fastq will be written to
+output.fastq with the adapter sequence removed. Adapter matching is
+error-tolerant. Multiple adapter sequences can be given (use further -a
+options), but only the best-matching adapter will be removed.
 
 Input may also be in FASTA format. Compressed input and output is supported and
 auto-detected from the file name (.gz, .xz, .bz2). Use the file name '-' for
@@ -55,23 +54,26 @@ See https://cutadapt.readthedocs.io/ for full documentation.
 """
 
 import sys
-import errno
 import time
-from argparse import ArgumentParser, SUPPRESS, HelpFormatter
 import logging
 import platform
-from xopen import xopen
+from typing import Tuple, Optional, Sequence, List, Any, Iterator, Union
+from argparse import ArgumentParser, SUPPRESS, HelpFormatter
+
 import dnaio
 
 from cutadapt import __version__
-from cutadapt.adapters import AdapterParser
-from cutadapt.modifiers import (LengthTagModifier, SuffixRemover, PrefixSuffixAdder,
+from cutadapt.adapters import warn_duplicate_adapters, Adapter
+from cutadapt.parser import AdapterParser
+from cutadapt.modifiers import (Modifier, LengthTagModifier, SuffixRemover, PrefixSuffixAdder,
     ZeroCapper, QualityTrimmer, UnconditionalCutter, NEndTrimmer, AdapterCutter,
-    PairedAdapterCutterError, PairedAdapterCutter, NextseqQualityTrimmer, Shortener)
+    PairedAdapterCutterError, PairedAdapterCutter, NextseqQualityTrimmer, Shortener,
+    ReverseComplementer)
 from cutadapt.report import full_report, minimal_report
-from cutadapt.pipeline import (SingleEndPipeline, PairedEndPipeline, InputFiles, OutputFiles,
-    SerialPipelineRunner, ParallelPipelineRunner)
-from cutadapt.utils import available_cpu_count
+from cutadapt.pipeline import (Pipeline, SingleEndPipeline, PairedEndPipeline, InputFiles,
+    OutputFiles, SerialPipelineRunner, ParallelPipelineRunner)
+from cutadapt.utils import available_cpu_count, Progress, DummyProgress, FileOpener
+from cutadapt.log import setup_logging, REPORT
 
 logger = logging.getLogger()
 
@@ -107,51 +109,8 @@ class CommandLineError(Exception):
     pass
 
 
-# Custom log level, see setup_logging
-REPORT = 25
-
-
-class NiceFormatter(logging.Formatter):
-    """
-    Do not prefix "INFO:" to info-level log messages (but do it for all other
-    levels).
-
-    Based on http://stackoverflow.com/a/9218261/715090 .
-    """
-    def format(self, record):
-        if record.levelno not in (logging.INFO, REPORT):
-            record.msg = '{}: {}'.format(record.levelname, record.msg)
-        return super().format(record)
-
-
-def setup_logging(stdout=False, minimal=False, quiet=False, debug=False):
-    """
-    Attach handler to the global logger object
-    """
-    # For --report=minimal, we need this custom log level because we want to
-    # print nothing except the minimal report and therefore cannot use the
-    # INFO level (and the ERROR level would give us an 'ERROR:' prefix).
-    logging.addLevelName(REPORT, 'REPORT')
-
-    # Due to backwards compatibility, logging output is sent to standard output
-    # instead of standard error if the -o option is used.
-    stream_handler = logging.StreamHandler(sys.stdout if stdout else sys.stderr)
-    stream_handler.setFormatter(NiceFormatter())
-    # debug overrides quiet overrides minimal
-    if debug:
-        level = logging.DEBUG
-    elif quiet:
-        level = logging.ERROR
-    elif minimal:
-        level = REPORT
-    else:
-        level = logging.INFO
-    stream_handler.setLevel(level)
-    logger.setLevel(level)
-    logger.addHandler(stream_handler)
-
-
-def get_argument_parser():
+def get_argument_parser() -> ArgumentParser:
+    # noqa: E131
     parser = CutadaptArgumentParser(usage=__doc__, add_help=False)
     group = parser.add_argument_group("Options")
     group.add_argument("-h", "--help", action="help", help="Show this help message and exit")
@@ -171,7 +130,7 @@ def get_argument_parser():
     group.add_argument("--buffer-size", type=int, default=4000000,
         help=SUPPRESS)
     # Compression level for gzipped output files. Not exposed since we have -Z
-    group.add_argument("--compression-level", default=6,
+    group.add_argument("--compression-level", type=int, default=6,
         help=SUPPRESS)
     # Deprecated: The input format is always auto-detected
     group.add_argument("-f", "--format", help=SUPPRESS)
@@ -183,19 +142,21 @@ def get_argument_parser():
             "trimmed (but see the --times option). When the special notation "
             "'file:FILE' is used, adapter sequences are read from the given "
             "FASTA file.")
-    group.add_argument("-a", "--adapter", action="append", default=[], metavar="ADAPTER",
-        dest="adapters",
+    group.add_argument("-a", "--adapter", type=lambda x: ("back", x), action="append",
+        default=[], metavar="ADAPTER", dest="adapters",
         help="Sequence of an adapter ligated to the 3' end (paired data: of the "
             "first read). The adapter and subsequent bases are trimmed. If a "
             "'$' character is appended ('anchoring'), the adapter is only "
             "found if it is a suffix of the read.")
-    group.add_argument("-g", "--front", action="append", default=[], metavar="ADAPTER",
+    group.add_argument("-g", "--front", type=lambda x: ("front", x), action="append",
+        default=[], metavar="ADAPTER", dest="adapters",
         help="Sequence of an adapter ligated to the 5' end (paired data: of the "
             "first read). The adapter and any preceding bases are trimmed. "
             "Partial matches at the 5' end are allowed. If a '^' character is "
             "prepended ('anchoring'), the adapter is only found if it is a "
             "prefix of the read.")
-    group.add_argument("-b", "--anywhere", action="append", default=[], metavar="ADAPTER",
+    group.add_argument("-b", "--anywhere", type=lambda x: ("anywhere", x), action="append",
+        default=[], metavar="ADAPTER", dest="adapters",
         help="Sequence of an adapter that may be ligated to the 5' or 3' end "
             "(paired data: of the first read). Both types of matches as "
             "described under -a und -g are allowed. If the first base of the "
@@ -224,7 +185,12 @@ def get_argument_parser():
             "mask: replace with 'N' characters; "
             "lowercase: convert to lowercase; "
             "none: leave unchanged (useful with "
-            "--discard-untrimmed). Default: trim")
+            "--discard-untrimmed). Default: %(default)s")
+    group.add_argument("--rc", "--revcomp", dest="reverse_complement", default=False,
+        action="store_true",
+        help="Check both the read and its reverse complement for adapter matches. If "
+            "match is on reverse-complemented version, output that one. "
+            "Default: check only read")
     group.add_argument("--no-trim", dest='action', action='store_const', const='none',
         help=SUPPRESS)  # Deprecated, use --action=none
     group.add_argument("--mask-adapter", dest='action', action='store_const', const='mask',
@@ -326,11 +292,14 @@ def get_argument_parser():
     group = parser.add_argument_group("Paired-end options", description="The "
         "-A/-G/-B/-U options work like their -a/-b/-g/-u counterparts, but "
         "are applied to the second read in each pair.")
-    group.add_argument("-A", dest='adapters2', action='append', default=[], metavar='ADAPTER',
+    group.add_argument("-A", type=lambda x: ("back", x), dest='adapters2',
+        action='append', default=[], metavar='ADAPTER',
         help="3' adapter to be removed from second read in a pair.")
-    group.add_argument("-G", dest='front2', action='append', default=[], metavar='ADAPTER',
+    group.add_argument("-G", type=lambda x: ("front", x), dest='adapters2',
+        action='append', default=[], metavar='ADAPTER',
         help="5' adapter to be removed from second read in a pair.")
-    group.add_argument("-B", dest='anywhere2', action='append', default=[], metavar='ADAPTER',
+    group.add_argument("-B", type=lambda x: ("anywhere", x), dest='adapters2',
+        action='append', default=[], metavar='ADAPTER',
         help="5'/3 adapter to be removed from second read in a pair.")
     group.add_argument("-U", dest='cut2', action='append', default=[], type=int, metavar="LENGTH",
         help="Remove LENGTH bases from second read in a pair.")
@@ -372,26 +341,29 @@ def get_argument_parser():
     return parser
 
 
-def parse_cutoffs(s):
-    """Parse a string INT[,INT] into a two-element list of integers"""
-    cutoffs = s.split(',')
+def parse_cutoffs(s: str) -> Tuple[int, int]:
+    """Parse a string INT[,INT] into a pair of integers
+
+    >>> parse_cutoffs("5")
+    (0, 5)
+    >>> parse_cutoffs("6,7")
+    (6, 7)
+    """
+    try:
+        cutoffs = [int(value) for value in s.split(",")]
+    except ValueError as e:
+        raise CommandLineError("Quality cutoff value not recognized: {}".format(e))
+
     if len(cutoffs) == 1:
-        try:
-            cutoffs = [0, int(cutoffs[0])]
-        except ValueError as e:
-            raise CommandLineError("Quality cutoff value not recognized: {}".format(e))
-    elif len(cutoffs) == 2:
-        try:
-            cutoffs = [int(cutoffs[0]), int(cutoffs[1])]
-        except ValueError as e:
-            raise CommandLineError("Quality cutoff value not recognized: {}".format(e))
-    else:
+        cutoffs = [0, cutoffs[0]]
+    elif len(cutoffs) != 2:
         raise CommandLineError("Expected one value or two values separated by comma for "
             "the quality cutoff")
-    return cutoffs
+
+    return (cutoffs[0], cutoffs[1])
 
 
-def parse_lengths(s):
+def parse_lengths(s: str) -> Tuple[Optional[int], ...]:
     """Parse [INT][:[INT]] into a pair of integers. If a value is omitted, use None
 
     >>> parse_lengths('25')
@@ -415,62 +387,38 @@ def parse_lengths(s):
     return tuple(values)
 
 
-def open_output_files(args, default_outfile, interleaved):
+def open_output_files(
+    args, default_outfile, interleaved: bool, file_opener: FileOpener
+) -> OutputFiles:
     """
     Return an OutputFiles instance. If demultiplex is True, the untrimmed, untrimmed2, out and out2
     attributes are not opened files, but paths (out and out2 with the '{name}' template).
     """
-    compression_level = args.compression_level
-    rest_file = info_file = wildcard = None
-    if args.rest_file is not None:
-        rest_file = xopen(args.rest_file, 'w', compresslevel=compression_level)
-    if args.info_file is not None:
-        info_file = xopen(args.info_file, 'w', compresslevel=compression_level)
-    if args.wildcard_file is not None:
-        wildcard = xopen(args.wildcard_file, 'w', compresslevel=compression_level)
 
-    def open2(path1, path2):
-        file1 = file2 = None
-        if path1 is not None:
-            file1 = xopen(path1, 'wb', compresslevel=compression_level)
-            if path2 is not None:
-                file2 = xopen(path2, 'wb', compresslevel=compression_level)
-        return file1, file2
+    rest_file = file_opener.xopen_or_none(args.rest_file, "wb")
+    info_file = file_opener.xopen_or_none(args.info_file, "wb")
+    wildcard = file_opener.xopen_or_none(args.wildcard_file, "wb")
 
     too_short = too_short2 = None
     if args.minimum_length is not None:
-        too_short, too_short2 = open2(args.too_short_output, args.too_short_paired_output)
+        too_short, too_short2 = file_opener.xopen_pair(
+            args.too_short_output, args.too_short_paired_output, "wb")
 
     too_long = too_long2 = None
     if args.maximum_length is not None:
-        too_long, too_long2 = open2(args.too_long_output, args.too_long_paired_output)
+        too_long, too_long2 = file_opener.xopen_pair(
+            args.too_long_output, args.too_long_paired_output, "wb")
 
     if int(args.discard_trimmed) + int(args.discard_untrimmed) + int(
             args.untrimmed_output is not None) > 1:
         raise CommandLineError("Only one of the --discard-trimmed, --discard-untrimmed "
             "and --untrimmed-output options can be used at the same time.")
 
-    demultiplex = args.output is not None and '{name}' in args.output
-
-    if args.paired_output is not None and (demultiplex != ('{name}' in args.paired_output)):
-        raise CommandLineError('When demultiplexing paired-end data, "{name}" must appear in '
-            'both output file names (-o and -p)')
-
-    demultiplex_combinatorial = (
-        args.output is not None
-        and args.paired_output is not None
-        and '{name1}' in args.output
-        and '{name2}' in args.output
-        and '{name1}' in args.paired_output
-        and '{name2}' in args.paired_output
-    )
-    if (demultiplex or demultiplex_combinatorial) and args.discard_trimmed:
+    demultiplex_mode = determine_demultiplex_mode(args)
+    if demultiplex_mode and args.discard_trimmed:
         raise CommandLineError("Do not use --discard-trimmed when demultiplexing.")
 
-    if demultiplex:
-        if demultiplex_combinatorial:
-            raise CommandLineError("You cannot combine {name} with {name1} and {name2}")
-
+    if demultiplex_mode == "normal":
         out = args.output
         untrimmed = args.output.replace('{name}', 'unknown')
         if args.untrimmed_output:
@@ -485,12 +433,11 @@ def open_output_files(args, default_outfile, interleaved):
                 untrimmed2 = args.untrimmed_paired_output
             if args.discard_untrimmed:
                 untrimmed2 = None
-
         else:
             untrimmed2 = out2 = None
 
         assert out is not None and '{name}' in out and (out2 is None or '{name}' in out2)
-    elif demultiplex_combinatorial:
+    elif demultiplex_mode == "combinatorial":
         out = args.output
         out2 = args.paired_output
         if args.untrimmed_output or args.untrimmed_paired_output:
@@ -501,8 +448,9 @@ def open_output_files(args, default_outfile, interleaved):
         else:
             untrimmed = untrimmed2 = 'unknown'
     else:
-        untrimmed, untrimmed2 = open2(args.untrimmed_output, args.untrimmed_paired_output)
-        out, out2 = open2(args.output, args.paired_output)
+        untrimmed, untrimmed2 = file_opener.xopen_pair(
+            args.untrimmed_output, args.untrimmed_paired_output, "wb")
+        out, out2 = file_opener.xopen_pair(args.output, args.paired_output, "wb")
         if out is None:
             out = default_outfile
 
@@ -518,13 +466,41 @@ def open_output_files(args, default_outfile, interleaved):
         untrimmed2=untrimmed2,
         out=out,
         out2=out2,
-        demultiplex=demultiplex or demultiplex_combinatorial,
+        demultiplex=bool(demultiplex_mode),
         interleaved=interleaved,
         force_fasta=args.fasta,
     )
 
 
-def determine_paired_mode(args):
+def determine_demultiplex_mode(args) -> Union[str, bool]:
+    """Return one of "normal", "combinatorial" or False"""
+
+    demultiplex = args.output is not None and '{name}' in args.output
+
+    if args.paired_output is not None and (demultiplex != ('{name}' in args.paired_output)):
+        raise CommandLineError('When demultiplexing paired-end data, "{name}" must appear in '
+                               'both output file names (-o and -p)')
+
+    demultiplex_combinatorial = (
+        args.output is not None
+        and args.paired_output is not None
+        and '{name1}' in args.output
+        and '{name2}' in args.output
+        and '{name1}' in args.paired_output
+        and '{name2}' in args.paired_output
+    )
+    if demultiplex and demultiplex_combinatorial:
+        raise CommandLineError("You cannot combine {name} with {name1} and {name2}")
+
+    if demultiplex:
+        return "normal"
+    elif demultiplex_combinatorial:
+        return "combinatorial"
+    else:
+        return False
+
+
+def determine_paired(args) -> bool:
     """
     Determine whether we should work in paired-end mode.
     """
@@ -533,15 +509,13 @@ def determine_paired_mode(args):
         args.paired_output
         or args.interleaved
         or args.adapters2
-        or args.front2
-        or args.anywhere2
         or args.cut2
         or args.pair_filter
         or args.too_short_paired_output
         or args.too_long_paired_output)
 
 
-def determine_interleaved(args):
+def determine_interleaved(args) -> Tuple[bool, bool]:
     is_interleaved_input = False
     is_interleaved_output = False
     if args.interleaved:
@@ -553,7 +527,9 @@ def determine_interleaved(args):
     return is_interleaved_input, is_interleaved_output
 
 
-def input_files_from_parsed_args(inputs, paired, interleaved):
+def setup_input_files(
+    inputs: Sequence[str], paired: bool, interleaved: bool
+) -> Tuple[str, Optional[str]]:
     """
     Return tuple (input_filename, input_paired_filename)
     """
@@ -576,7 +552,7 @@ def input_files_from_parsed_args(inputs, paired, interleaved):
                 "but then you also need to provide two input files (you provided one) or "
                 "use --interleaved.")
         else:
-            input_paired_filename = inputs[1]
+            input_paired_filename = inputs[1]  # type: Optional[str]
     else:
         if len(inputs) == 2:
             raise CommandLineError(
@@ -588,19 +564,15 @@ def input_files_from_parsed_args(inputs, paired, interleaved):
     return input_filename, input_paired_filename
 
 
-def pipeline_from_parsed_args(args, paired, is_interleaved_output):
-    """
-    Setup a processing pipeline from parsed command-line arguments.
-
-    If there are any problems parsing the arguments, a CommandLineError is thrown.
-
-    Return an instance of Pipeline (SingleEndPipeline or PairedEndPipeline)
-    """
-
+def check_arguments(args, paired: bool, is_interleaved_output: bool) -> None:
     if not paired:
         if args.untrimmed_paired_output:
             raise CommandLineError("Option --untrimmed-paired-output can only be used when "
-                "trimming paired-end reads (with option -p).")
+                "trimming paired-end reads.")
+
+        if args.pair_adapters:
+            raise CommandLineError("Option --pair-adapters can only be used when trimming "
+                "paired-end reads")
 
     if paired:
         if not is_interleaved_output:
@@ -631,62 +603,40 @@ def pipeline_from_parsed_args(args, paired, is_interleaved_output):
         raise CommandLineError("The overlap must be at least 1.")
     if not (0 <= args.gc_content <= 100):
         raise CommandLineError("GC content must be given as percentage between 0 and 100")
+
+    if args.pair_adapters and args.times != 1:
+        raise CommandLineError("--pair-adapters cannot be used with --times")
+
+
+def pipeline_from_parsed_args(args, paired, file_opener) -> Pipeline:
+    """
+    Setup a processing pipeline from parsed command-line arguments.
+
+    If there are any problems parsing the arguments, a CommandLineError is raised.
+
+    Return an instance of Pipeline (SingleEndPipeline or PairedEndPipeline)
+    """
     if args.action == 'none':
         args.action = None
 
-    adapter_parser = AdapterParser(
-        max_error_rate=args.error_rate,
-        min_overlap=args.overlap,
-        read_wildcards=args.match_read_wildcards,
-        adapter_wildcards=args.match_adapter_wildcards,
-        indels=args.indels,
-    )
-    try:
-        adapters = adapter_parser.parse_multi(args.adapters, args.anywhere, args.front)
-        adapters2 = adapter_parser.parse_multi(args.adapters2, args.anywhere2, args.front2)
-    except IOError as e:
-        if e.errno == errno.ENOENT:
-            raise CommandLineError(e)
-        raise
-    except ValueError as e:
-        raise CommandLineError(e)
-    if args.debug:
-        for adapter in adapters + adapters2:
-            adapter.enable_debug()
+    adapters, adapters2 = adapters_from_args(args)
 
     # Create the processing pipeline
     if paired:
         pair_filter_mode = 'any' if args.pair_filter is None else args.pair_filter
-        pipeline = PairedEndPipeline(pair_filter_mode)
+        pipeline = PairedEndPipeline(
+            pair_filter_mode, file_opener
+        )  # type: Any
     else:
-        pipeline = SingleEndPipeline()
+        pipeline = SingleEndPipeline(file_opener)
 
     # When adapters are being trimmed only in R1 or R2, override the pair filter mode
     # as using the default of 'any' would regard all read pairs as untrimmed.
-    if paired and (not adapters2 or not adapters) and (
+    if isinstance(pipeline, PairedEndPipeline) and (not adapters2 or not adapters) and (
             args.discard_untrimmed or args.untrimmed_output or args.untrimmed_paired_output):
         pipeline.override_untrimmed_pair_filter = True
 
-    for i, cut_arg in enumerate([args.cut, args.cut2]):
-        # cut_arg is a list
-        if not cut_arg:
-            continue
-        if len(cut_arg) > 2:
-            raise CommandLineError("You cannot remove bases from more than two ends.")
-        if len(cut_arg) == 2 and cut_arg[0] * cut_arg[1] > 0:
-            raise CommandLineError("You cannot remove bases from the same end twice.")
-        for c in cut_arg:
-            if c == 0:
-                continue
-            if i == 0:  # R1
-                if paired:
-                    pipeline.add(UnconditionalCutter(c), None)
-                else:
-                    pipeline.add(UnconditionalCutter(c))
-            else:
-                # R2
-                assert isinstance(pipeline, PairedEndPipeline)
-                pipeline.add(None, UnconditionalCutter(c))
+    add_unconditional_cutters(pipeline, args.cut, args.cut2, paired)
 
     pipeline_add = pipeline.add_both if paired else pipeline.add
 
@@ -696,43 +646,19 @@ def pipeline_from_parsed_args(args, paired, is_interleaved_output):
         cutoffs = parse_cutoffs(args.quality_cutoff)
         pipeline_add(QualityTrimmer(cutoffs[0], cutoffs[1], args.quality_base))
 
-    if args.pair_adapters:
-        if not paired:
-            raise CommandLineError("Option --pair-adapters can only be used when trimming "
-                "paired-end reads")
-        if args.times != 1:
-            raise CommandLineError("--pair-adapters cannot be used with --times")
-        try:
-            cutter = PairedAdapterCutter(adapters, adapters2, args.action)
-        except PairedAdapterCutterError as e:
-            raise CommandLineError("--pair-adapters: " + str(e))
-        pipeline.add_paired_modifier(cutter)
-    else:
-        adapter_cutter, adapter_cutter2 = None, None
-        if adapters:
-            adapter_cutter = AdapterCutter(adapters, args.times, args.action)
-        if adapters2:
-            adapter_cutter2 = AdapterCutter(adapters2, args.times, args.action)
-        if paired:
-            if adapter_cutter or adapter_cutter2:
-                pipeline.add(adapter_cutter, adapter_cutter2)
-        else:
-            if adapter_cutter:
-                pipeline.add(adapter_cutter)
+    add_adapter_cutter(
+        pipeline,
+        adapters,
+        adapters2,
+        paired,
+        args.pair_adapters,
+        args.action,
+        args.times,
+        args.reverse_complement,
+    )
 
-    # Remaining modifiers that apply to both reads of paired-end reads
-    if args.length is not None:
-        pipeline_add(Shortener(args.length))
-    if args.trim_n:
-        pipeline_add(NEndTrimmer())
-    if args.length_tag:
-        pipeline_add(LengthTagModifier(args.length_tag))
-    for suffix in args.strip_suffix:
-        pipeline_add(SuffixRemover(suffix))
-    if args.prefix or args.suffix:
-        pipeline_add(PrefixSuffixAdder(args.prefix, args.suffix))
-    if args.zero_cap:
-        pipeline_add(ZeroCapper(quality_base=args.quality_base))
+    for modifier in modifiers_applying_to_both_ends_if_paired(args):
+        pipeline_add(modifier)
 
     # Set filtering parameters
     # Minimum/maximum length
@@ -753,6 +679,106 @@ def pipeline_from_parsed_args(args, paired, is_interleaved_output):
     return pipeline
 
 
+def adapters_from_args(args) -> Tuple[List[Adapter], List[Adapter]]:
+    adapter_parser = AdapterParser(
+        max_error_rate=args.error_rate,
+        min_overlap=args.overlap,
+        read_wildcards=args.match_read_wildcards,
+        adapter_wildcards=args.match_adapter_wildcards,
+        indels=args.indels,
+    )
+    try:
+        adapters = adapter_parser.parse_multi(args.adapters)
+        adapters2 = adapter_parser.parse_multi(args.adapters2)
+    except (FileNotFoundError, ValueError) as e:
+        raise CommandLineError(e)
+    warn_duplicate_adapters(adapters)
+    warn_duplicate_adapters(adapters2)
+    if args.debug:
+        for adapter in adapters + adapters2:
+            adapter.enable_debug()
+    return adapters, adapters2
+
+
+def add_unconditional_cutters(pipeline: Pipeline, cut1: List[int], cut2: List[int], paired: bool):
+    for i, cut_arg in enumerate([cut1, cut2]):
+        # cut_arg is a list
+        if not cut_arg:
+            continue
+        if len(cut_arg) > 2:
+            raise CommandLineError("You cannot remove bases from more than two ends.")
+        if len(cut_arg) == 2 and cut_arg[0] * cut_arg[1] > 0:
+            raise CommandLineError("You cannot remove bases from the same end twice.")
+        for c in cut_arg:
+            if c == 0:
+                continue
+            if i == 0:  # R1
+                if paired:
+                    assert isinstance(pipeline, PairedEndPipeline)
+                    pipeline.add(UnconditionalCutter(c), None)
+                else:
+                    assert isinstance(pipeline, SingleEndPipeline)
+                    pipeline.add(UnconditionalCutter(c))
+            else:
+                # R2
+                assert isinstance(pipeline, PairedEndPipeline)
+                pipeline.add(None, UnconditionalCutter(c))
+
+
+def add_adapter_cutter(
+    pipeline,
+    adapters,
+    adapters2,
+    paired: bool,
+    pair_adapters: bool,
+    action: str,
+    times: int,
+    reverse_complement: bool,
+):
+    if pair_adapters:
+        if reverse_complement:
+            raise CommandLineError("Cannot use --revcomp with --pair-adapters")
+        try:
+            cutter = PairedAdapterCutter(adapters, adapters2, action)
+        except PairedAdapterCutterError as e:
+            raise CommandLineError("--pair-adapters: " + str(e))
+        pipeline.add_paired_modifier(cutter)
+    else:
+        adapter_cutter, adapter_cutter2 = None, None
+        if adapters:
+            adapter_cutter = AdapterCutter(adapters, times, action)
+        if adapters2:
+            adapter_cutter2 = AdapterCutter(adapters2, times, action)
+        if paired:
+            if reverse_complement:
+                raise CommandLineError("--revcomp not implemented for paired-end reads")
+            if adapter_cutter or adapter_cutter2:
+                pipeline.add(adapter_cutter, adapter_cutter2)
+        elif adapter_cutter:
+            if reverse_complement:
+                modifier = ReverseComplementer(
+                    adapter_cutter
+                )  # type: Union[AdapterCutter,ReverseComplementer]
+            else:
+                modifier = adapter_cutter
+            pipeline.add(modifier)
+
+
+def modifiers_applying_to_both_ends_if_paired(args) -> Iterator[Modifier]:
+    if args.length is not None:
+        yield Shortener(args.length)
+    if args.trim_n:
+        yield NEndTrimmer()
+    if args.length_tag:
+        yield LengthTagModifier(args.length_tag)
+    for suffix in args.strip_suffix:
+        yield SuffixRemover(suffix)
+    if args.prefix or args.suffix:
+        yield PrefixSuffixAdder(args.prefix, args.suffix)
+    if args.zero_cap:
+        yield ZeroCapper(quality_base=args.quality_base)
+
+
 def log_header(cmdlineargs):
     """Print the "This is cutadapt ..." header"""
 
@@ -780,12 +806,14 @@ def main(cmdlineargs=None, default_outfile=sys.stdout.buffer):
     # Setup logging only if there are not already any handlers (can happen when
     # this function is being called externally such as from unit tests)
     if not logging.root.handlers:
-        setup_logging(stdout=log_to_stdout,
+        setup_logging(logger, stdout=log_to_stdout,
             quiet=args.quiet, minimal=args.report == 'minimal', debug=args.debug)
     if args.profile:
         import cProfile
         profiler = cProfile.Profile()
         profiler.enable()
+    else:
+        profiler = None
 
     if args.quiet and args.report:
         parser.error("Options --quiet and --report cannot be used at the same time")
@@ -797,46 +825,48 @@ def main(cmdlineargs=None, default_outfile=sys.stdout.buffer):
             "--strip-f3, --maq, --bwa, --no-zero-cap. "
             "Use Cutadapt 1.18 or earlier to work with colorspace data.")
 
-    paired = determine_paired_mode(args)
+    paired = determine_paired(args)
     assert paired in (False, True)
 
     # Print the header now because some of the functions below create logging output
     log_header(cmdlineargs)
-    try:
-        is_interleaved_input, is_interleaved_output = determine_interleaved(args)
-        input_filename, input_paired_filename = input_files_from_parsed_args(args.inputs,
-            paired, is_interleaved_input)
-        pipeline = pipeline_from_parsed_args(args, paired, is_interleaved_output)
-        outfiles = open_output_files(args, default_outfile, is_interleaved_output)
-    except CommandLineError as e:
-        parser.error(e)
-        return  # avoid IDE warnings below
-
     if args.cores < 0:
         parser.error('Value for --cores cannot be negative')
+
     cores = available_cpu_count() if args.cores == 0 else args.cores
+    file_opener = FileOpener(
+        compression_level=args.compression_level, threads=0 if cores == 1 else None)
+    try:
+        is_interleaved_input, is_interleaved_output = determine_interleaved(args)
+        input_filename, input_paired_filename = setup_input_files(args.inputs,
+            paired, is_interleaved_input)
+        check_arguments(args, paired, is_interleaved_output)
+        pipeline = pipeline_from_parsed_args(args, paired, file_opener)
+        outfiles = open_output_files(args, default_outfile, is_interleaved_output, file_opener)
+    except CommandLineError as e:
+        parser.error(str(e))
+        return  # avoid IDE warnings below
+
     if cores > 1:
         if ParallelPipelineRunner.can_output_to(outfiles):
             runner_class = ParallelPipelineRunner
             runner_kwargs = dict(n_workers=cores, buffer_size=args.buffer_size)
         else:
-            logger.error('Running in parallel is currently not supported for '
-                'the given combination of command-line parameters.\nThese '
-                'options are not supported: --info-file, --rest-file, '
-                '--wildcard-file, --untrimmed-output, '
-                '--untrimmed-paired-output, --too-short-output, '
-                '--too-short-paired-output, --too-long-output, '
-                '--too-long-paired-output, --format\n'
-                'Also, demultiplexing is not supported.\n'
-                'Omit --cores/-j to continue.')
+            parser.error("Running in parallel is currently not supported "
+                "when using --format or when demultiplexing.\n"
+                "Omit --cores/-j to continue.")
             return  # avoid IDE warnings below
     else:
         runner_class = SerialPipelineRunner
         runner_kwargs = dict()
     infiles = InputFiles(input_filename, file2=input_paired_filename,
             interleaved=is_interleaved_input)
+    if sys.stderr.isatty() and not args.quiet:
+        progress = Progress()
+    else:
+        progress = DummyProgress()
     try:
-        runner = runner_class(pipeline, infiles, outfiles, **runner_kwargs)
+        runner = runner_class(pipeline, infiles, outfiles, progress, **runner_kwargs)
     except (dnaio.UnknownFileFormat, IOError) as e:
         parser.error(e)
         return  # avoid IDE warnings below
@@ -850,10 +880,8 @@ def main(cmdlineargs=None, default_outfile=sys.stdout.buffer):
     except KeyboardInterrupt:
         print("Interrupted", file=sys.stderr)
         sys.exit(130)
-    except IOError as e:
-        if e.errno == errno.EPIPE:
-            sys.exit(1)
-        raise
+    except BrokenPipeError:
+        sys.exit(1)
     except (dnaio.FileFormatError, dnaio.UnknownFileFormat, EOFError) as e:
         sys.exit("cutadapt: error: {}".format(e))
 
@@ -863,7 +891,7 @@ def main(cmdlineargs=None, default_outfile=sys.stdout.buffer):
     else:
         report = full_report
     logger.log(REPORT, '%s', report(stats, elapsed, args.gc_content / 100))
-    if args.profile:
+    if profiler is not None:
         import pstats
         profiler.disable()
         pstats.Stats(profiler).sort_stats('time').print_stats(20)
