@@ -4,14 +4,13 @@ Adapter finding and trimming classes
 The ...Adapter classes are responsible for finding adapters.
 The ...Match classes trim the reads.
 """
-import re
 import logging
 from enum import Enum
 from collections import defaultdict
-from dnaio.readers import FastaReader
+from typing import Optional, Tuple, Sequence, Dict, Any, List
+from abc import ABC, abstractmethod
 
-from cutadapt import align
-from .align import hamming_environment
+from . import align
 
 logger = logging.getLogger()
 
@@ -31,418 +30,34 @@ class Where(Enum):
     LINKED = 'linked'
 
 
-# TODO put this in some kind of "list of pre-defined adapter types" along with the info above
+class WhereToRemove(Enum):
+    PREFIX = 1
+    SUFFIX = 2
+    AUTO = 3
+
+
 WHERE_TO_REMOVE_MAP = {
-    Where.PREFIX: 'prefix',
-    Where.FRONT_NOT_INTERNAL: 'prefix',
-    Where.FRONT: 'prefix',
-    Where.BACK: 'suffix',
-    Where.SUFFIX: 'suffix',
-    Where.BACK_NOT_INTERNAL: 'suffix',
-    Where.ANYWHERE: 'auto',
+    Where.PREFIX: WhereToRemove.PREFIX,
+    Where.FRONT_NOT_INTERNAL: WhereToRemove.PREFIX,
+    Where.FRONT: WhereToRemove.PREFIX,
+    Where.BACK: WhereToRemove.SUFFIX,
+    Where.SUFFIX: WhereToRemove.SUFFIX,
+    Where.BACK_NOT_INTERNAL: WhereToRemove.SUFFIX,
+    Where.ANYWHERE: WhereToRemove.AUTO,
 }
 
 
-class AdapterSpecification:
-    """
-    Description of a single adapter. This represents the same information that would be
-    given on the commandline through an option such as -a, but does not handle linked
-    adapters on also not the file: syntax.
-
-    These are the attributes:
-
-    - name (None or str)
-    - restriction (None, 'anchored', or 'noninternal')
-    - sequence (nucleotide sequence as string)
-    - parameters (dict with extra parameters such as 'max_error_rate', 'min_overlap')
-    - cmdline_type ('front' for -a, 'back' for -g and 'anywhere' for -b)
-
-    >>> AdapterSpecification.parse('a_name=ACGT;anywhere', 'back')
-    AdapterSpecification(name='a_name', restriction=None, sequence='ACGT', parameters={'anywhere': True}, cmdline_type='back')
-    """
-
-    def __init__(self, name, restriction, sequence, parameters, cmdline_type):
-        self.name = name
-        self.restriction = restriction
-        self.sequence = sequence
-        self.parameters = parameters
-        self.cmdline_type = cmdline_type
-
-    @classmethod
-    def parse(cls, spec: str, cmdline_type: str):
-        """Factory for creating an instance from a string specification"""
-        name, restriction, sequence, parameters = cls._parse(spec, cmdline_type)
-        return cls(name, restriction, sequence, parameters, cmdline_type)
-
-    def __repr__(self):
-        return '{}(name={!r}, restriction={!r}, sequence={!r}, parameters={!r}, cmdline_type={!r})'.format(
-            self.__class__.__name__, self.name, self.restriction, self.sequence, self.parameters, self.cmdline_type)
-
-    def __eq__(self, other):
-        return (
-            self.name == other.name
-            and self.restriction == other.restriction
-            and self.sequence == other.sequence
-            and self.parameters == other.parameters
-            and self.cmdline_type == other.cmdline_type
-        )
-
-    @staticmethod
-    def expand_braces(sequence):
-        """
-        Replace all occurrences of ``x{n}`` (where x is any character) with n
-        occurrences of x. Raise ValueError if the expression cannot be parsed.
-
-        >>> AdapterSpecification.expand_braces('TGA{5}CT')
-        'TGAAAAACT'
-        """
-        # Simple DFA with four states, encoded in prev
-        result = ''
-        prev = None
-        for s in re.split('([{}])', sequence):
-            if s == '':
-                continue
-            if prev is None:
-                if s == '{':
-                    raise ValueError('"{" must be used after a character')
-                if s == '}':
-                    raise ValueError('"}" cannot be used here')
-                prev = s
-                result += s
-            elif prev == '{':
-                prev = int(s)
-                if not 0 <= prev <= 10000:
-                    raise ValueError('Value {} invalid'.format(prev))
-            elif isinstance(prev, int):
-                if s != '}':
-                    raise ValueError('"}" expected')
-                result = result[:-1] + result[-1] * prev
-                prev = None
-            else:
-                if s != '{':
-                    raise ValueError('Expected "{"')
-                prev = '{'
-        # Check if we are in a non-terminating state
-        if isinstance(prev, int) or prev == '{':
-            raise ValueError("Unterminated expression")
-        return result
-
-    @staticmethod
-    def _extract_name(spec):
-        """
-        Parse an adapter specification given as 'name=adapt' into 'name' and 'adapt'.
-        """
-        fields = spec.split('=', 1)
-        if len(fields) > 1:
-            name, spec = fields
-            name = name.strip()
-        else:
-            name = None
-        spec = spec.strip()
-        return name, spec
-
-    allowed_parameters = {
-        # abbreviations
-        'e': 'max_error_rate',
-        'error_rate': 'max_error_rate',
-        'o': 'min_overlap',
-
-        # allowed parameters
-        'max_error_rate': None,
-        'min_overlap': None,
-        'anywhere': None,
-        'required': None,
-        'optional': None,  # If this is specified, 'required' will be set to False
-    }
-
-    @classmethod
-    def _parse_parameters(cls, spec):
-        """Parse key=value;key=value;key=value into a dict"""
-
-        fields = spec.split(';')
-        result = dict()
-        for field in fields:
-            field = field.strip()
-            if not field:
-                continue
-            key, equals, value = field.partition('=')
-            if equals == '=' and value == '':
-                raise ValueError('No value given')
-            key = key.strip()
-            if key not in cls.allowed_parameters:
-                raise KeyError('Unknown parameter {}'.format(key))
-            # unabbreviate
-            while cls.allowed_parameters[key] is not None:
-                key = cls.allowed_parameters[key]
-            value = value.strip()
-            if value == '':
-                value = True
-            else:
-                try:
-                    value = int(value)
-                except ValueError:
-                    value = float(value)
-            if key in result:
-                raise KeyError('Key {} specified twice'.format(key))
-            result[key] = value
-        if 'optional' in result and 'required' in result:
-            raise ValueError("'optional' and 'required' cannot be specified at the same time")
-        if 'optional' in result:
-            result['required'] = False
-            del result['optional']
-        return result
-
-    @classmethod
-    def _parse(cls, spec, cmdline_type):
-        """
-        Parse an adapter specification for a non-linked adapter (without '...')
-
-        Allow:
-        'back' and ADAPTER
-        'back' and ADAPTERX
-        'back' and ADAPTER$
-        'front' and ADAPTER
-        'front' and XADAPTER
-        'front' and ^ADAPTER
-        'anywhere' and ADAPTER
-        """
-        if cmdline_type not in ("front", "back", "anywhere"):
-            raise ValueError("cmdline_type must be front, back or anywhere")
-        error = ValueError(
-            "You cannot use multiple placement restrictions for an adapter at the same time. "
-            "Choose one of ^ADAPTER, ADAPTER$, XADAPTER or ADAPTERX")
-        spec, middle, parameters_spec = spec.partition(';')
-        name, spec = cls._extract_name(spec)
-        spec = spec.strip()
-        parameters = cls._parse_parameters(parameters_spec)
-        spec = AdapterSpecification.expand_braces(spec)
-
-        # Special case for adapters consisting of only X characters:
-        # This needs to be supported for backwards-compatibilitity
-        if len(spec.strip('X')) == 0:
-            return name, None, spec, {}
-
-        front_restriction = None
-        if spec.startswith('^'):
-            front_restriction = 'anchored'
-            spec = spec[1:]
-        if spec.upper().startswith('X'):
-            if front_restriction is not None:
-                raise error
-            front_restriction = 'noninternal'
-            spec = spec.lstrip('xX')
-
-        back_restriction = None
-        if spec.endswith('$'):
-            back_restriction = 'anchored'
-            spec = spec[:-1]
-        if spec.upper().endswith('X'):
-            if back_restriction is not None:
-                raise error
-            back_restriction = 'noninternal'
-            spec = spec.rstrip('xX')
-
-        n_placement_restrictions = int(bool(front_restriction)) + int(bool(back_restriction))
-        if n_placement_restrictions > 1:
-            raise error
-
-        if cmdline_type == 'front' and back_restriction:
-            raise ValueError(
-                "Allowed placement restrictions for a 5' adapter are XADAPTER and ^ADAPTER")
-        if cmdline_type == 'back' and front_restriction:
-            raise ValueError(
-                "Allowed placement restrictions for a 3' adapter are ADAPTERX and ADAPTER$")
-
-        assert front_restriction is None or back_restriction is None
-        if front_restriction is not None:
-            restriction = front_restriction
-        else:
-            restriction = back_restriction
-
-        if cmdline_type == 'anywhere' and restriction is not None:
-            raise ValueError(
-                "Placement restrictions (with X, ^, $) not supported for 'anywhere' (-b) adapters")
-
-        return name, restriction, spec, parameters
-
-    @staticmethod
-    def _restriction_to_where(cmdline_type, restriction):
-        if cmdline_type == 'front':
-            if restriction is None:
-                return Where.FRONT
-            elif restriction == 'anchored':
-                return Where.PREFIX
-            elif restriction == 'noninternal':
-                return Where.FRONT_NOT_INTERNAL
-            else:
-                raise ValueError(
-                    'Value {} for a front restriction not allowed'.format(restriction))
-        elif cmdline_type == 'back':
-            if restriction is None:
-                return Where.BACK
-            elif restriction == 'anchored':
-                return Where.SUFFIX
-            elif restriction == 'noninternal':
-                return Where.BACK_NOT_INTERNAL
-            else:
-                raise ValueError(
-                    'Value {} for a back restriction not allowed'.format(restriction))
-        else:
-            assert cmdline_type == 'anywhere'
-            if restriction is None:
-                return Where.ANYWHERE
-            else:
-                raise ValueError('No placement may be specified for "anywhere" adapters')
-
-    def where(self):
-        return self._restriction_to_where(self.cmdline_type, self.restriction)
-
-
-class AdapterParser:
-    """
-    Factory for Adapter classes that all use the same default parameters (error rate,
-    indels etc.). The given **kwargs will be passed to the Adapter constructors.
-    """
-    def __init__(self, **kwargs):
-        # kwargs: max_error_rate, min_overlap, read_wildcards, adapter_wildcards, indels
-        self.default_parameters = kwargs
-
-    def _parse(self, spec: str, cmdline_type='back', name=None):
-        """
-        Parse an adapter specification not using ``file:`` notation and return
-        an object of an appropriate Adapter class.
-
-        name -- Adapter name if not included as part of the spec. (If spec is
-        'name=ADAPTER', name will be 'name'.)
-
-        cmdline_type -- describes which commandline parameter was used (``-a``
-        is 'back', ``-b`` is 'anywhere', and ``-g`` is 'front').
-        """
-        if cmdline_type not in ('front', 'back', 'anywhere'):
-            raise ValueError('cmdline_type cannot be {!r}'.format(cmdline_type))
-        spec1, middle, spec2 = spec.partition('...')
-        del spec
-
-        # Handle linked adapter
-        if middle == '...' and spec1 and spec2:
-            return self._parse_linked(spec1, spec2, name, cmdline_type)
-        elif middle == '...':
-            if not spec1:
-                if cmdline_type == 'back':  # -a ...ADAPTER
-                    spec = spec2
-                else:  # -g ...ADAPTER
-                    raise ValueError('Invalid adapter specification')
-            elif not spec2:
-                if cmdline_type == 'back':  # -a ADAPTER...
-                    cmdline_type = 'front'
-                    spec = spec1
-                else:  # -g ADAPTER...
-                    spec = spec1
-            else:
-                assert False, 'This should not happen'
-        else:
-            spec = spec1
-
-        spec = AdapterSpecification.parse(spec, cmdline_type)
-        where = spec.where()
-        if not name:
-            name = spec.name
-        if spec.parameters.pop('anywhere', False):
-            spec.parameters['remove'] = WHERE_TO_REMOVE_MAP[where]
-            where = Where.ANYWHERE
-        parameters = self.default_parameters.copy()
-        parameters.update(spec.parameters)
-        if where in (Where.FRONT, Where.BACK):
-            adapter_class = BackOrFrontAdapter
-        else:
-            adapter_class = Adapter
-        return adapter_class(sequence=spec.sequence, where=where, name=name, **parameters)
-
-    def _parse_linked(self, spec1: str, spec2: str, name, cmdline_type):
-        """Return a linked adapter from two specification strings"""
-
-        if cmdline_type == 'anywhere':
-            raise ValueError("'anywhere' (-b) adapters may not be linked")
-        front_spec = AdapterSpecification.parse(spec1, 'front')
-        back_spec = AdapterSpecification.parse(spec2, 'back')
-        if not name:
-            name = front_spec.name
-
-        if cmdline_type == 'back' and front_spec.restriction is None:
-            import textwrap
-            logger.warning('\n'.join(textwrap.wrap(
-                "You specified a linked adapter as '-a ADAPTER1...ADAPTER2'. "
-                "The interpretation of what this means has changed in Cutadapt 2.0. "
-                "(The 5' adapter is now no longer anchored by default.) "
-                "To get results consist with the old behavior, you need to anchor "
-                "the 5' adapter explicitly as in '-a ^ADAPTER1...ADAPTER2'."
-            )))
-
-        front_anchored = front_spec.restriction is not None
-        back_anchored = back_spec.restriction is not None
-
-        front_parameters = self.default_parameters.copy()
-        front_parameters.update(front_spec.parameters)
-        back_parameters = self.default_parameters.copy()
-        back_parameters.update(back_spec.parameters)
-
-        if cmdline_type == 'front':
-            # -g requires both adapters to be present
-            front_required = True
-            back_required = True
-        else:
-            # -a requires only the anchored adapters to be present
-            front_required = front_anchored
-            back_required = back_anchored
-
-        # Handle parameters overriding whether an adapter is required
-        front_required = front_spec.parameters.pop('required', front_required)
-        back_required = back_spec.parameters.pop('required', back_required)
-
-        front_adapter = Adapter(front_spec.sequence, where=front_spec.where(), name=None,
-            **front_spec.parameters)
-        back_adapter = Adapter(back_spec.sequence, where=back_spec.where(), name=None,
-            **back_spec.parameters)
-
-        return LinkedAdapter(
-            front_adapter=front_adapter,
-            back_adapter=back_adapter,
-            front_required=front_required,
-            back_required=back_required,
-            name=name,
-        )
-
-    def parse(self, spec, cmdline_type='back'):
-        """
-        Parse an adapter specification and yield appropriate Adapter classes.
-        This works like the _parse_no_file() function above, but also supports the
-        ``file:`` notation for reading adapters from an external FASTA
-        file. Since a file can contain multiple adapters, this
-        function is a generator.
-        """
-        if spec.startswith('file:'):
-            # read adapter sequences from a file
-            with FastaReader(spec[5:]) as fasta:
-                for record in fasta:
-                    name = record.name.split(None, 1)
-                    name = name[0] if name else None
-                    yield self._parse(record.sequence, cmdline_type, name=name)
-        else:
-            yield self._parse(spec, cmdline_type, name=None)
-
-    def parse_multi(self, back, anywhere, front):
-        """
-        Parse all three types of commandline options that can be used to
-        specify adapters. back, anywhere and front are lists of strings,
-        corresponding to the respective commandline types (-a, -b, -g).
-
-        Return a list of appropriate Adapter classes.
-        """
-        adapters = []
-        for specs, cmdline_type in (back, 'back'), (anywhere, 'anywhere'), (front, 'front'):
-            for spec in specs:
-                adapters.extend(self.parse(spec, cmdline_type))
-        return adapters
+# TODO could become a property/attribute of the Adapter classes
+ADAPTER_TYPE_NAMES = {
+    Where.BACK: "regular 3'",
+    Where.BACK_NOT_INTERNAL: "non-internal 3'",
+    Where.FRONT: "regular 5'",
+    Where.FRONT_NOT_INTERNAL: "non-internal 5'",
+    Where.PREFIX: "anchored 5'",
+    Where.SUFFIX: "anchored 3'",
+    Where.ANYWHERE: "variable 5'/3'",
+    Where.LINKED: "linked",
+}
 
 
 def returns_defaultdict_int():
@@ -455,18 +70,27 @@ def returns_defaultdict_int():
 class EndStatistics:
     """Statistics about the 5' or 3' end"""
 
-    def __init__(self, adapter):
-        self.where = adapter.where
-        self.max_error_rate = adapter.max_error_rate
-        self.sequence = adapter.sequence
-        self.effective_length = adapter.effective_length
-        self.has_wildcards = adapter.adapter_wildcards
+    def __init__(self, adapter: "SingleAdapter"):
+        self.where = adapter.where  # type: Where
+        self.max_error_rate = adapter.max_error_rate  # type: float
+        self.sequence = adapter.sequence  # type: str
+        self.effective_length = adapter.effective_length  # type: int
+        self.has_wildcards = adapter.adapter_wildcards  # type: bool
         # self.errors[l][e] == n iff n times a sequence of length l matching at e errors was removed
-        self.errors = defaultdict(returns_defaultdict_int)
-        self._remove = adapter.remove
+        self.errors = defaultdict(returns_defaultdict_int)  # type: Dict[int, Dict[int, int]]
+        self._remove = adapter.remove  # type: Optional[WhereToRemove]
         self.adjacent_bases = {'A': 0, 'C': 0, 'G': 0, 'T': 0, '': 0}
 
-    def __iadd__(self, other):
+    def __repr__(self):
+        errors = {k: dict(v) for k, v in self.errors.items()}
+        return "EndStatistics(where={}, max_error_rate={}, errors={}, adjacent_bases={})".format(
+            self.where,
+            self.max_error_rate,
+            errors,
+            self.adjacent_bases,
+        )
+
+    def __iadd__(self, other: Any):
         if not isinstance(other, self.__class__):
             raise ValueError("Cannot compare")
         if (
@@ -503,16 +127,16 @@ class EndStatistics:
         """
         seq = self.sequence
         # FIXME this is broken for self._remove == 'auto'
-        if self._remove == 'prefix':
+        if self._remove == WhereToRemove.PREFIX:
             seq = seq[::-1]
         allowed_bases = 'CGRYSKMBDHVN' if self.has_wildcards else 'GC'
-        p = 1
+        p = 1.
         probabilities = [p]
         for i, c in enumerate(seq):
             if c in allowed_bases:
                 p *= gc_content / 2.
             else:
-                p *= (1 - gc_content) / 2
+                p *= (1. - gc_content) / 2.
             probabilities.append(p)
         return probabilities
 
@@ -521,27 +145,52 @@ class AdapterStatistics:
     """
     Statistics about an adapter. An adapter can work on the 5' end (front)
     or 3' end (back) of a read, and statistics for that are captured
-    separately.
+    separately in EndStatistics objects.
     """
 
-    def __init__(self, adapter, adapter2=None, where=None):
+    def __init__(
+        self,
+        adapter: "SingleAdapter",
+        adapter2: Optional["SingleAdapter"] = None,
+        where: Optional[Where] = None,
+    ):
         self.name = adapter.name
         self.where = where if where is not None else adapter.where
         self.front = EndStatistics(adapter)
+        self.reverse_complemented = 0
         if adapter2 is None:
             self.back = EndStatistics(adapter)
         else:
             self.back = EndStatistics(adapter2)
 
-    def __iadd__(self, other):
+    def __repr__(self):
+        return "AdapterStatistics(name={}, where={}, front={}, back={})".format(
+            self.name,
+            self.where,
+            self.front,
+            self.back,
+        )
+
+    def __iadd__(self, other: "AdapterStatistics"):
         if self.where != other.where:  # TODO self.name != other.name or
             raise ValueError('incompatible objects')
         self.front += other.front
         self.back += other.back
+        self.reverse_complemented += other.reverse_complemented
         return self
 
 
-class Match:
+class Match(ABC):
+    @abstractmethod
+    def remainder_interval(self) -> Tuple[int, int]:
+        pass
+
+    @abstractmethod
+    def get_info_records(self) -> List[List]:
+        pass
+
+
+class SingleMatch(Match):
     """
     Representation of a single adapter matched to a single read.
     """
@@ -549,37 +198,48 @@ class Match:
         'adapter', 'read', 'length', '_trimmed_read', 'adjacent_base']
 
     # TODO Can remove_before be removed from the constructor parameters?
-    def __init__(self, astart, astop, rstart, rstop, matches, errors, remove_before, adapter, read):
+    def __init__(
+        self,
+        astart: int,
+        astop: int,
+        rstart: int,
+        rstop: int,
+        matches: int,
+        errors: int,
+        remove_before: bool,
+        adapter: "SingleAdapter",
+        read,
+    ):
         """
         remove_before -- True: remove bases before adapter. False: remove after
         """
-        self.astart = astart
-        self.astop = astop
-        self.rstart = rstart
-        self.rstop = rstop
-        self.matches = matches
-        self.errors = errors
-        self.adapter = adapter
+        self.astart = astart  # type: int
+        self.astop = astop  # type: int
+        self.rstart = rstart  # type: int
+        self.rstop = rstop  # type: int
+        self.matches = matches  # type: int
+        self.errors = errors  # type: int
+        self.adapter = adapter  # type: SingleAdapter
         self.read = read
         if remove_before:
             # Compute the trimmed read, assuming it’s a 'front' adapter
             self._trimmed_read = read[rstop:]
-            self.adjacent_base = ''
+            self.adjacent_base = ""  # type: str
         else:
             # Compute the trimmed read, assuming it’s a 'back' adapter
             self.adjacent_base = read.sequence[rstart - 1:rstart]
             self._trimmed_read = read[:rstart]
-        self.remove_before = remove_before
+        self.remove_before = remove_before  # type: bool
         # Number of aligned characters in the adapter. If there are
         # indels, this may be different from the number of characters
         # in the read.
-        self.length = astop - astart
+        self.length = astop - astart  # type: int
 
     def __repr__(self):
-        return 'Match(astart={}, astop={}, rstart={}, rstop={}, matches={}, errors={})'.format(
+        return 'SingleMatch(astart={}, astop={}, rstart={}, rstop={}, matches={}, errors={})'.format(
             self.astart, self.astop, self.rstart, self.rstop, self.matches, self.errors)
 
-    def wildcards(self, wildcard_char='N'):
+    def wildcards(self, wildcard_char: str = "N") -> str:
         """
         Return a string that contains, for each wildcard character,
         the character that it matches. For example, if the adapter
@@ -593,7 +253,7 @@ class Match:
                 self.rstart + i < len(self.read.sequence)]
         return ''.join(wildcards)
 
-    def rest(self):
+    def rest(self) -> str:
         """
         Return the part of the read before this match if this is a
         'front' (5') adapter,
@@ -605,10 +265,20 @@ class Match:
         else:
             return self.read.sequence[self.rstop:]
 
-    def get_info_record(self):
+    def remainder_interval(self) -> Tuple[int, int]:
+        """
+        Return an interval (start, stop) that describes the part of the read that would
+        remain after trimming
+        """
+        if self.remove_before:
+            return self.rstop, len(self.read.sequence)
+        else:
+            return 0, self.rstart
+
+    def get_info_records(self) -> List[List]:
         seq = self.read.sequence
         qualities = self.read.qualities
-        info = (
+        info = [
             self.read.name,
             self.errors,
             self.rstart,
@@ -616,23 +286,23 @@ class Match:
             seq[0:self.rstart],
             seq[self.rstart:self.rstop],
             seq[self.rstop:],
-            self.adapter.name
-        )
+            self.adapter.name,
+        ]
         if qualities:
-            info += (
+            info += [
                 qualities[0:self.rstart],
                 qualities[self.rstart:self.rstop],
                 qualities[self.rstop:]
-            )
+            ]
         else:
-            info += ('', '', '')
+            info += ["", "", ""]
 
-        return info
+        return [info]
 
     def trimmed(self):
         return self._trimmed_read
 
-    def update_statistics(self, statistics):
+    def update_statistics(self, statistics: AdapterStatistics):
         """Update AdapterStatistics in place"""
         if self.remove_before:
             statistics.front.errors[self.rstop][self.errors] += 1
@@ -644,13 +314,26 @@ class Match:
                 statistics.back.adjacent_bases[''] = 1
 
 
-def _generate_adapter_name(_start=[1]):
+def _generate_adapter_name(_start=[1]) -> str:
     name = str(_start[0])
     _start[0] += 1
     return name
 
 
-class Adapter:
+class Adapter(ABC):
+    def __init__(self, *args, **kwargs):
+        pass
+
+    @abstractmethod
+    def enable_debug(self):
+        pass
+
+    @abstractmethod
+    def match_to(self, read):
+        pass
+
+
+class SingleAdapter(Adapter):
     """
     This class can find a single adapter characterized by sequence, error rate,
     type etc. within reads.
@@ -658,11 +341,12 @@ class Adapter:
     where --  A Where enum value. This influences where the adapter is allowed to appear within the
         read.
 
-    remove -- describes which part of the read to remove if the adapter was found:
-          * "prefix" (for a 3' adapter)
-          * "suffix" (for a 5' adapter)
-          * "auto" for a 5'/3' mixed adapter (if the match involves the first base of the read, it
-            is assumed to be a 5' adapter and a 3' otherwise)
+    remove -- a WhereToRemove enum value. This describes which part of the read to remove if the
+        adapter was found:
+          * WhereToRemove.PREFIX (for a 3' adapter)
+          * WhereToRemove.SUFFIX (for a 5' adapter)
+          * WhereToRemove.AUTO for a 5'/3' mixed adapter (if the match involves the first base of
+            the read, it is assumed to be a 5' adapter and a 3' otherwise)
           * None: One of the above is chosen depending on the 'where' parameter
 
     sequence -- The adapter sequence as string. Will be converted to uppercase.
@@ -684,19 +368,28 @@ class Adapter:
         unique number.
     """
 
-    def __init__(self, sequence, where, remove=None, max_error_rate=0.1, min_overlap=3,
-            read_wildcards=False, adapter_wildcards=True, name=None, indels=True):
-        self._debug = False
-        self.name = _generate_adapter_name() if name is None else name
-        self.sequence = sequence.upper().replace('U', 'T')
+    def __init__(
+        self,
+        sequence: str,
+        where: Where,
+        remove: Optional[WhereToRemove] = None,
+        max_error_rate: float = 0.1,
+        min_overlap: int = 3,
+        read_wildcards: bool = False,
+        adapter_wildcards: bool = True,
+        name: Optional[str] = None,
+        indels: bool = True,
+    ):
+        super().__init__()
+        self._debug = False  # type: bool
+        self.name = _generate_adapter_name() if name is None else name  # type: str
+        self.sequence = sequence.upper().replace("U", "T")  # type: str
         if not self.sequence:
-            raise ValueError('Sequence is empty')
-        self.where = where
-        if remove not in (None, 'prefix', 'suffix', 'auto'):
-            raise ValueError('remove parameter must be "prefix", "suffix", "auto" or None')
-        self.remove = WHERE_TO_REMOVE_MAP[where] if remove is None else remove
-        self.max_error_rate = max_error_rate
-        self.min_overlap = min(min_overlap, len(self.sequence))
+            raise ValueError("Adapter sequence is empty")
+        self.where = where  # type: Where
+        self.remove = WHERE_TO_REMOVE_MAP[where] if remove is None else remove  # type: WhereToRemove
+        self.max_error_rate = max_error_rate  # type: float
+        self.min_overlap = min(min_overlap, len(self.sequence))  # type: int
         iupac = frozenset('XACGTURYSWKMBDHVN')
         if adapter_wildcards and not set(self.sequence) <= iupac:
             for c in self.sequence:
@@ -705,9 +398,9 @@ class Adapter:
                         'not a valid IUPAC code. Use only characters '
                         'XACGTURYSWKMBDHVN.'.format(c, self.sequence))
         # Optimization: Use non-wildcard matching if only ACGT is used
-        self.adapter_wildcards = adapter_wildcards and not set(self.sequence) <= set('ACGT')
-        self.read_wildcards = read_wildcards
-        self.indels = indels
+        self.adapter_wildcards = adapter_wildcards and not set(self.sequence) <= set("ACGT")  # type: bool
+        self.read_wildcards = read_wildcards  # type: bool
+        self.indels = indels  # type: bool
         if self.is_anchored and not self.indels:
             aligner_class = align.PrefixComparer if self.where is Where.PREFIX else align.SuffixComparer
             self.aligner = aligner_class(
@@ -733,7 +426,7 @@ class Adapter:
             )
 
     def __repr__(self):
-        return '<Adapter(name={name!r}, sequence={sequence!r}, where={where}, '\
+        return '<SingleAdapter(name={name!r}, sequence={sequence!r}, where={where}, '\
             'remove={remove}, max_error_rate={max_error_rate}, min_overlap={min_overlap}, '\
             'read_wildcards={read_wildcards}, '\
             'adapter_wildcards={adapter_wildcards}, '\
@@ -796,12 +489,12 @@ class Adapter:
 
         if match_args is None:
             return None
-        if self.remove == 'auto':
+        if self.remove == WhereToRemove.AUTO:
             # guess: if alignment starts at pos 0, it’s a 5' adapter
             remove_before = match_args[2] == 0  # index 2 is rstart
         else:
-            remove_before = self.remove == 'prefix'
-        match = Match(*match_args, remove_before=remove_before, adapter=self, read=read)
+            remove_before = self.remove == WhereToRemove.PREFIX
+        match = SingleMatch(*match_args, remove_before=remove_before, adapter=self, read=read)
 
         assert match.length >= self.min_overlap
         return match
@@ -813,7 +506,7 @@ class Adapter:
         return AdapterStatistics(self)
 
 
-class BackOrFrontAdapter(Adapter):
+class BackOrFrontAdapter(SingleAdapter):
     """A 5' or 3' adapter.
 
     This is separate from the Adapter class so that a specialized match_to
@@ -826,7 +519,7 @@ class BackOrFrontAdapter(Adapter):
     def __init__(self, *args, **kwargs):
         super().__init__(*args, **kwargs)
         assert self.where is Where.BACK or self.where is Where.FRONT
-        self._remove_before = self.remove == 'prefix'
+        self._remove_before = self.remove is WhereToRemove.PREFIX
 
     def match_to(self, read):
         """
@@ -853,21 +546,19 @@ class BackOrFrontAdapter(Adapter):
         if alignment is None:
             return None
 
-        match = Match(*alignment, remove_before=self._remove_before, adapter=self, read=read)
+        match = SingleMatch(*alignment, remove_before=self._remove_before, adapter=self, read=read)
         return match
 
 
-class LinkedMatch:
+class LinkedMatch(Match):
     """
     Represent a match of a LinkedAdapter
     """
-    def __init__(self, front_match, back_match, adapter):
-        """
-        One of front_match and back_match must be not None!
-        """
-        self.front_match = front_match
-        self.back_match = back_match
-        self.adapter = adapter
+    def __init__(self, front_match: SingleMatch, back_match: SingleMatch, adapter: "LinkedAdapter"):
+        assert front_match is not None or back_match is not None
+        self.front_match = front_match  # type: SingleMatch
+        self.back_match = back_match  # type: SingleMatch
+        self.adapter = adapter  # type: LinkedAdapter
 
     def __repr__(self):
         return '<LinkedMatch(front_match={!r}, back_match={}, adapter={})>'.format(
@@ -876,11 +567,22 @@ class LinkedMatch:
     @property
     def matches(self):
         """Number of matching bases"""
-        m = getattr(self.front_match, 'matches', 0)
+        m = 0
+        if self.front_match is not None:
+            m += self.front_match.matches
         if self.back_match is not None:
             m += self.back_match.matches
         return m
 
+    @property
+    def errors(self):
+        e = 0
+        if self.front_match is not None:
+            e += self.front_match.errors
+        if self.back_match is not None:
+            e += self.back_match.errors
+        return e
+
     def trimmed(self):
         if self.back_match:
             # back match is relative to front match, so even if a front match exists,
@@ -901,8 +603,25 @@ class LinkedMatch:
         if self.back_match:
             statistics.back.errors[len(self.back_match.read) - self.back_match.rstart][self.back_match.errors] += 1
 
+    def remainder_interval(self) -> Tuple[int, int]:
+        matches = [match for match in [self.front_match, self.back_match] if match is not None]
+        return remainder(matches)
 
-class LinkedAdapter:
+    def get_info_records(self) -> List[List]:
+        records = []
+        for match, namesuffix in [
+            (self.front_match, ";1"),
+            (self.back_match, ";2"),
+        ]:
+            if match is None:
+                continue
+            record = match.get_info_records()[0]
+            record[7] = self.adapter.name + namesuffix
+            records.append(record)
+        return records
+
+
+class LinkedAdapter(Adapter):
     """
     """
     def __init__(
@@ -913,6 +632,7 @@ class LinkedAdapter:
         back_required,
         name,
     ):
+        super().__init__()
         self.front_required = front_required
         self.back_required = back_required
 
@@ -920,6 +640,7 @@ class LinkedAdapter:
         self.where = Where.LINKED
         self.name = _generate_adapter_name() if name is None else name
         self.front_adapter = front_adapter
+        self.front_adapter.name = self.name
         self.back_adapter = back_adapter
 
     def enable_debug(self):
@@ -947,12 +668,20 @@ class LinkedAdapter:
     def create_statistics(self):
         return AdapterStatistics(self.front_adapter, self.back_adapter, where=Where.LINKED)
 
+    @property
+    def sequence(self):
+        return self.front_adapter.sequence + "..." + self.back_adapter.sequence
 
-class MultiAdapter:
+    @property
+    def remove(self):
+        return None
+
+
+class MultiAdapter(Adapter):
     """
     Represent multiple adapters of the same type at once and use an index data structure
     to speed up matching. This acts like a "normal" Adapter as it provides a match_to
-    method, but is faster with lots of adaptes.
+    method, but is faster with lots of adapters.
 
     There are quite a few restrictions:
     - the adapters need to be either all PREFIX or all SUFFIX adapters
@@ -966,6 +695,7 @@ class MultiAdapter:
 
     def __init__(self, adapters):
         """All given adapters must be of the same type, either Where.PREFIX or Where.SUFFIX"""
+        super().__init__()
         if not adapters:
             raise ValueError("Adapter list is empty")
         self._where = adapters[0].where
@@ -1016,7 +746,7 @@ class MultiAdapter:
         for adapter in self._adapters:
             sequence = adapter.sequence
             k = int(adapter.max_error_rate * len(sequence))
-            for s, errors, matches in hamming_environment(sequence, k):
+            for s, errors, matches in align.hamming_environment(sequence, k):
                 if s in index:
                     other_adapter, other_errors, other_matches = index[s]
                     if matches < other_matches:
@@ -1079,14 +809,45 @@ class MultiAdapter:
             else:
                 assert self._where is Where.SUFFIX
                 rstart, rstop = len(read) - best_length, len(read)
-            return Match(
+            return SingleMatch(
                 astart=0,
                 astop=len(best_adapter.sequence),
                 rstart=rstart,
                 rstop=rstop,
                 matches=best_m,
                 errors=best_e,
-                remove_before=best_adapter.remove == 'prefix',
+                remove_before=best_adapter.remove is WhereToRemove.PREFIX,
                 adapter=best_adapter,
                 read=read
             )
+
+    def enable_debug(self):
+        pass
+
+
+def warn_duplicate_adapters(adapters):
+    d = dict()
+    for adapter in adapters:
+        key = (adapter.sequence, adapter.where, adapter.remove)
+        if key in d:
+            logger.warning("Adapter %r (%s) was specified multiple times! "
+                "Please make sure that this is what you want.",
+                adapter.sequence, ADAPTER_TYPE_NAMES[adapter.where])
+        d[key] = adapter.name
+
+
+def remainder(matches: Sequence[Match]) -> Tuple[int, int]:
+    """
+    Determine which section of the read would not be trimmed. Return a tuple (start, stop)
+    that gives the interval of the untrimmed part relative to the original read.
+
+    matches must be non-empty
+    """
+    if not matches:
+        raise ValueError("matches must not be empty")
+    start = 0
+    for match in matches:
+        match_start, match_stop = match.remainder_interval()
+        start += match_start
+    length = match_stop - match_start
+    return (start, start + length)
