@@ -4,16 +4,19 @@ import sys
 import copy
 import logging
 import functools
+from typing import List, IO, Optional, BinaryIO, TextIO, Any, Tuple, Union
 from abc import ABC, abstractmethod
 from multiprocessing import Process, Pipe, Queue
+from pathlib import Path
 import multiprocessing.connection
+from multiprocessing.connection import Connection
 import traceback
 
 from xopen import xopen
 import dnaio
 
-from .utils import Progress
-from .modifiers import PairedModifier, Modifier
+from .utils import Progress, FileOpener
+from .modifiers import Modifier, PairedModifier, PairedModifierWrapper
 from .report import Statistics
 from .filters import (Redirector, PairedRedirector, NoFilter, PairedNoFilter, InfoFileWriter,
     RestFileWriter, WildcardFileWriter, TooShortReadFilter, TooLongReadFilter, NContentFilter,
@@ -24,7 +27,7 @@ logger = logging.getLogger()
 
 
 class InputFiles:
-    def __init__(self, file1, file2=None, interleaved=False):
+    def __init__(self, file1: BinaryIO, file2: Optional[BinaryIO] = None, interleaved: bool = False):
         self.file1 = file1
         self.file2 = file2
         self.interleaved = interleaved
@@ -35,26 +38,28 @@ class OutputFiles:
     The attributes are open file-like objects except when demultiplex is True. In that case,
     untrimmed, untrimmed2, out and out2 are file names or templates
     as required by the used demultiplexer ('{name}' etc.).
+
     If interleaved is True, then out is written interleaved.
+
     Files may also be None.
     """
     # TODO interleaving for the other file pairs (too_short, too_long, untrimmed)?
     def __init__(
-            self,
-            out=None,
-            out2=None,
-            untrimmed=None,
-            untrimmed2=None,
-            too_short=None,
-            too_short2=None,
-            too_long=None,
-            too_long2=None,
-            info=None,
-            rest=None,
-            wildcard=None,
-            demultiplex=False,
-            interleaved=False,
-            force_fasta=None,
+        self,
+        out: Optional[BinaryIO] = None,
+        out2: Optional[BinaryIO] = None,
+        untrimmed: Optional[BinaryIO] = None,
+        untrimmed2: Optional[BinaryIO] = None,
+        too_short: Optional[BinaryIO] = None,
+        too_short2: Optional[BinaryIO] = None,
+        too_long: Optional[BinaryIO] = None,
+        too_long2: Optional[BinaryIO] = None,
+        info: Optional[BinaryIO] = None,
+        rest: Optional[BinaryIO] = None,
+        wildcard: Optional[BinaryIO] = None,
+        demultiplex: bool = False,
+        interleaved: bool = False,
+        force_fasta: Optional[bool] = None,
     ):
         self.out = out
         self.out2 = out2
@@ -91,13 +96,15 @@ class Pipeline(ABC):
     """
     n_adapters = 0
 
-    def __init__(self):
-        self._close_files = []
+    def __init__(self, file_opener: FileOpener):
+        self._close_files = []  # type: List[IO]
         self._reader = None
-        self._filters = []
-        self._modifiers = []
-        self._outfiles = None
+        self._filters = []  # type: List[Any]
+        # TODO type should be Union[List[Modifier], List[PairedModifier]]
+        self._modifiers = []  # type: List[Union[Modifier, PairedModifier]]
+        self._outfiles = None  # type: Optional[OutputFiles]
         self._demultiplexer = None
+        self._textiowrappers = []  # type: List[TextIO]
 
         # Filter settings
         self._minimum_length = None
@@ -106,46 +113,60 @@ class Pipeline(ABC):
         self.discard_casava = False
         self.discard_trimmed = False
         self.discard_untrimmed = False
+        self.file_opener = file_opener
 
-    def set_input(self, infiles: InputFiles):
+    def connect_io(self, infiles: InputFiles, outfiles: OutputFiles):
         self._reader = dnaio.open(infiles.file1, file2=infiles.file2,
             interleaved=infiles.interleaved, mode='r')
+        self._set_output(outfiles)
 
-    def _open_writer(self, file, file2, force_fasta=None, **kwargs):
+    def _open_writer(
+        self,
+        file: BinaryIO,
+        file2: Optional[BinaryIO] = None,
+        force_fasta: Optional[bool] = None,
+        **kwargs
+    ):
         # TODO backwards-incompatible change (?) would be to use outfiles.interleaved
         # for all outputs
         if force_fasta:
             kwargs['fileformat'] = 'fasta'
+        # file and file2 must already be file-like objects because we don’t want to
+        # take care of threads and compression levels here.
+        for f in (file, file2):
+            assert not isinstance(f, (str, bytes, Path))
         return dnaio.open(file, file2=file2, mode='w', qualities=self.uses_qualities,
             **kwargs)
 
-    def set_output(self, outfiles: OutputFiles):
+    def _set_output(self, outfiles: OutputFiles):
         self._filters = []
         self._outfiles = outfiles
         filter_wrapper = self._filter_wrapper()
 
-        if outfiles.rest:
-            self._filters.append(RestFileWriter(outfiles.rest))
-        if outfiles.info:
-            self._filters.append(InfoFileWriter(outfiles.info))
-        if outfiles.wildcard:
-            self._filters.append(WildcardFileWriter(outfiles.wildcard))
+        for filter_class, outfile in (
+            (RestFileWriter, outfiles.rest),
+            (InfoFileWriter, outfiles.info),
+            (WildcardFileWriter, outfiles.wildcard),
+        ):
+            if outfile:
+                textiowrapper = io.TextIOWrapper(outfile)
+                self._textiowrappers.append(textiowrapper)
+                self._filters.append(filter_wrapper(None, filter_class(textiowrapper), None))
 
         # minimum length and maximum length
         for lengths, file1, file2, filter_class in (
                 (self._minimum_length, outfiles.too_short, outfiles.too_short2, TooShortReadFilter),
                 (self._maximum_length, outfiles.too_long, outfiles.too_long2, TooLongReadFilter)
         ):
-            writer = None
-            if lengths is not None:
-                if file1:
-                    writer = self._open_writer(file1, file2)
-                f1 = filter_class(lengths[0]) if lengths[0] is not None else None
-                if len(lengths) == 2 and lengths[1] is not None:
-                    f2 = filter_class(lengths[1])
-                else:
-                    f2 = None
-                self._filters.append(filter_wrapper(writer, filter=f1, filter2=f2))
+            if lengths is None:
+                continue
+            writer = self._open_writer(file1, file2) if file1 else None
+            f1 = filter_class(lengths[0]) if lengths[0] is not None else None
+            if len(lengths) == 2 and lengths[1] is not None:
+                f2 = filter_class(lengths[1])
+            else:
+                f2 = None
+            self._filters.append(filter_wrapper(writer, filter=f1, filter2=f2))
 
         if self.max_n is not None:
             f1 = f2 = NContentFilter(self.max_n)
@@ -170,15 +191,29 @@ class Pipeline(ABC):
             # --discard-untrimmed and --untrimmed-output. These options
             # are mutually exclusive in order to avoid brain damage.
             if self.discard_trimmed:
-                self._filters.append(filter_wrapper(None, DiscardTrimmedFilter(), DiscardTrimmedFilter()))
+                self._filters.append(
+                    filter_wrapper(None, DiscardTrimmedFilter(), DiscardTrimmedFilter()))
             elif self.discard_untrimmed:
-                self._filters.append(untrimmed_filter_wrapper(None, DiscardUntrimmedFilter(), DiscardUntrimmedFilter()))
+                self._filters.append(
+                    untrimmed_filter_wrapper(None, DiscardUntrimmedFilter(), DiscardUntrimmedFilter()))
             elif outfiles.untrimmed:
                 untrimmed_writer = self._open_writer(outfiles.untrimmed, outfiles.untrimmed2)
-                self._filters.append(untrimmed_filter_wrapper(untrimmed_writer, DiscardUntrimmedFilter(), DiscardUntrimmedFilter()))
+                self._filters.append(
+                    untrimmed_filter_wrapper(untrimmed_writer, DiscardUntrimmedFilter(), DiscardUntrimmedFilter()))
             self._filters.append(self._final_filter(outfiles))
 
-    def close(self):
+    def flush(self) -> None:
+        for f in self._textiowrappers:
+            f.flush()
+        assert self._outfiles is not None
+        for f in self._outfiles:
+            if f is not None:
+                f.flush()
+
+    def close(self) -> None:
+        for f in self._textiowrappers:
+            f.close()  # This also closes the underlying files; a second close occurs below
+        assert self._outfiles is not None
         for f in self._outfiles:
             # TODO do not use hasattr
             if f is not None and f is not sys.stdin and f is not sys.stdout and hasattr(f, 'close'):
@@ -187,11 +222,11 @@ class Pipeline(ABC):
             self._demultiplexer.close()
 
     @property
-    def uses_qualities(self):
+    def uses_qualities(self) -> bool:
+        assert self._reader is not None
         return self._reader.delivers_qualities
 
     @abstractmethod
-    # TODO progress shouldn’t be a parameter
     def process_reads(self, progress: Progress = None):
         pass
 
@@ -204,11 +239,11 @@ class Pipeline(ABC):
         pass
 
     @abstractmethod
-    def _final_filter(self, outfiles):
+    def _final_filter(self, outfiles: OutputFiles):
         pass
 
     @abstractmethod
-    def _create_demultiplexer(self, outfiles):
+    def _create_demultiplexer(self, outfiles: OutputFiles):
         pass
 
 
@@ -218,11 +253,11 @@ class SingleEndPipeline(Pipeline):
     """
     paired = False
 
-    def __init__(self):
-        super().__init__()
+    def __init__(self, file_opener: FileOpener):
+        super().__init__(file_opener)
         self._modifiers = []
 
-    def add(self, modifier):
+    def add(self, modifier: Modifier):
         if modifier is None:
             raise ValueError("Modifier must not be None")
         self._modifiers.append(modifier)
@@ -231,6 +266,7 @@ class SingleEndPipeline(Pipeline):
         """Run the pipeline. Return statistics"""
         n = 0  # no. of processed reads  # TODO turn into attribute
         total_bp = 0
+        assert self._reader is not None
         for read in self._reader:
             n += 1
             if n % 10000 == 0 and progress:
@@ -250,12 +286,14 @@ class SingleEndPipeline(Pipeline):
     def _untrimmed_filter_wrapper(self):
         return Redirector
 
-    def _final_filter(self, outfiles):
-        writer = self._open_writer(outfiles.out, outfiles.out2, force_fasta=outfiles.force_fasta)
+    def _final_filter(self, outfiles: OutputFiles):
+        assert outfiles.out2 is None and outfiles.out is not None
+        writer = self._open_writer(outfiles.out, force_fasta=outfiles.force_fasta)
         return NoFilter(writer)
 
-    def _create_demultiplexer(self, outfiles):
-        return Demultiplexer(outfiles.out, outfiles.untrimmed, qualities=self.uses_qualities)
+    def _create_demultiplexer(self, outfiles: OutputFiles):
+        return Demultiplexer(outfiles.out, outfiles.untrimmed, qualities=self.uses_qualities,
+            file_opener=self.file_opener)
 
     @property
     def minimum_length(self):
@@ -282,40 +320,38 @@ class PairedEndPipeline(Pipeline):
     """
     paired = True
 
-    def __init__(self, pair_filter_mode):
-        """
-        Setting modify_first_read_only to True enables "legacy mode"
-        """
-        super().__init__()
+    def __init__(self, pair_filter_mode, file_opener: FileOpener):
+        super().__init__(file_opener)
         self._pair_filter_mode = pair_filter_mode
         self._reader = None
-        # Whether to gnore pair_filter mode for discard-untrimmed filter
+        # Whether to ignore pair_filter mode for discard-untrimmed filter
         self.override_untrimmed_pair_filter = False
 
-    def add(self, modifier1, modifier2):
+    def add(self, modifier1: Optional[Modifier], modifier2: Optional[Modifier]):
         """
         Add a modifier for R1 and R2. One of them can be None, in which case the modifier
         will only be added for the respective read.
         """
         if modifier1 is None and modifier2 is None:
             raise ValueError("Not both modifiers can be None")
-        self._modifiers.append(PairedModifier(modifier1, modifier2))
+        self._modifiers.append(PairedModifierWrapper(modifier1, modifier2))
 
-    def add_both(self, modifier):
+    def add_both(self, modifier: Modifier):
         """
         Add one modifier for both R1 and R2
         """
         assert modifier is not None
-        self._modifiers.append(PairedModifier(modifier, copy.copy(modifier)))
+        self._modifiers.append(PairedModifierWrapper(modifier, copy.copy(modifier)))
 
-    def add_paired_modifier(self, paired_modifier):
-        """Add a Modifier without wrapping it in a PairedModifier"""
+    def add_paired_modifier(self, paired_modifier: PairedModifier):
+        """Add a Modifier (without wrapping it in a PairedModifierWrapper)"""
         self._modifiers.append(paired_modifier)
 
     def process_reads(self, progress: Progress = None):
         n = 0  # no. of processed reads
         total1_bp = 0
         total2_bp = 0
+        assert self._reader is not None
         for read1, read2 in self._reader:
             n += 1
             if n % 10000 == 0 and progress:
@@ -356,10 +392,11 @@ class PairedEndPipeline(Pipeline):
     def _create_demultiplexer(self, outfiles):
         if '{name1}' in outfiles.out and '{name2}' in outfiles.out:
             return CombinatorialDemultiplexer(outfiles.out, outfiles.out2,
-                outfiles.untrimmed, qualities=self.uses_qualities)
+                outfiles.untrimmed, qualities=self.uses_qualities, file_opener=self.file_opener)
         else:
             return PairedDemultiplexer(outfiles.out, outfiles.out2,
-            outfiles.untrimmed, outfiles.untrimmed2, qualities=self.uses_qualities)
+                outfiles.untrimmed, outfiles.untrimmed2, qualities=self.uses_qualities,
+                file_opener=self.file_opener)
 
     @property
     def minimum_length(self):
@@ -380,53 +417,65 @@ class PairedEndPipeline(Pipeline):
         self._maximum_length = value
 
 
-def reader_process(file, file2, connections, queue, buffer_size, stdin_fd):
+class ReaderProcess(Process):
     """
-    Read chunks of FASTA or FASTQ data from *file* and send to a worker.
+    Read chunks of FASTA or FASTQ data (single-end or paired) and send to a worker.
 
-    queue -- a Queue of worker indices. A worker writes its own index into this
-        queue to notify the reader that it is ready to receive more data.
-    connections -- a list of Connection objects, one for each worker.
+    The reader repeatedly
 
-    The function repeatedly
-
-    - reads a chunk from the file
+    - reads a chunk from the file(s)
     - reads a worker index from the Queue
     - sends the chunk to connections[index]
 
-    and finally sends "poison pills" (the value -1) to all connections.
+    and finally sends the stop token -1 ("poison pills") to all connections.
     """
-    if stdin_fd != -1:
-        sys.stdin.close()
-        sys.stdin = os.fdopen(stdin_fd)
-    try:
-        with xopen(file, 'rb') as f:
-            if file2:
-                with xopen(file2, 'rb') as f2:
-                    for chunk_index, (chunk1, chunk2) in enumerate(dnaio.read_paired_chunks(f, f2, buffer_size)):
-                        # Determine the worker that should get this chunk
-                        worker_index = queue.get()
-                        pipe = connections[worker_index]
-                        pipe.send(chunk_index)
-                        pipe.send_bytes(chunk1)
-                        pipe.send_bytes(chunk2)
-            else:
-                for chunk_index, chunk in enumerate(dnaio.read_chunks(f, buffer_size)):
-                    # Determine the worker that should get this chunk
-                    worker_index = queue.get()
-                    pipe = connections[worker_index]
-                    pipe.send(chunk_index)
-                    pipe.send_bytes(chunk)
 
-        # Send poison pills to all workers
-        for _ in range(len(connections)):
-            worker_index = queue.get()
-            connections[worker_index].send(-1)
-    except Exception as e:
-        # TODO better send this to a common "something went wrong" Queue
-        for worker_index in range(len(connections)):
-            connections[worker_index].send(-2)
-            connections[worker_index].send((e, traceback.format_exc()))
+    def __init__(self, file, file2, connections, queue, buffer_size, stdin_fd):
+        """
+        queue -- a Queue of worker indices. A worker writes its own index into this
+            queue to notify the reader that it is ready to receive more data.
+        connections -- a list of Connection objects, one for each worker.
+        """
+        super().__init__()
+        self.file = file
+        self.file2 = file2
+        self.connections = connections
+        self.queue = queue
+        self.buffer_size = buffer_size
+        self.stdin_fd = stdin_fd
+
+    def run(self):
+        if self.stdin_fd != -1:
+            sys.stdin.close()
+            sys.stdin = os.fdopen(self.stdin_fd)
+        try:
+            with xopen(self.file, 'rb') as f:
+                if self.file2:
+                    with xopen(self.file2, 'rb') as f2:
+                        for chunk_index, (chunk1, chunk2) in enumerate(
+                                dnaio.read_paired_chunks(f, f2, self.buffer_size)):
+                            self.send_to_worker(chunk_index, chunk1, chunk2)
+                else:
+                    for chunk_index, chunk in enumerate(dnaio.read_chunks(f, self.buffer_size)):
+                        self.send_to_worker(chunk_index, chunk)
+
+            # Send poison pills to all workers
+            for _ in range(len(self.connections)):
+                worker_index = self.queue.get()
+                self.connections[worker_index].send(-1)
+        except Exception as e:
+            # TODO better send this to a common "something went wrong" Queue
+            for connection in self.connections:
+                connection.send(-2)
+                connection.send((e, traceback.format_exc()))
+
+    def send_to_worker(self, chunk_index, chunk1, chunk2=None):
+        worker_index = self.queue.get()
+        connection = self.connections[worker_index]
+        connection.send(chunk_index)
+        connection.send_bytes(chunk1)
+        if chunk2 is not None:
+            connection.send_bytes(chunk2)
 
 
 class WorkerProcess(Process):
@@ -437,13 +486,12 @@ class WorkerProcess(Process):
     To notify the reader process that it wants data, it puts its own identifier into the
     need_work_queue before attempting to read data from the read_pipe.
     """
-    def __init__(self, id_, pipeline, input_path1, input_path2,
+    def __init__(self, id_, pipeline, two_input_files,
             interleaved_input, orig_outfiles, read_pipe, write_pipe, need_work_queue):
         super().__init__()
         self._id = id_
         self._pipeline = pipeline
-        self._input_path1 = input_path1
-        self._input_path2 = input_path2
+        self._two_input_files = two_input_files
         self._interleaved_input = interleaved_input
         self._orig_outfiles = orig_outfiles
         self._read_pipe = read_pipe
@@ -466,45 +514,14 @@ class WorkerProcess(Process):
                     logger.error('%s', tb_str)
                     raise e
 
-                # Setting the .buffer.name attributess below is necessary because
-                # file format detection uses the file name
-                data = self._read_pipe.recv_bytes()
-                input = io.BytesIO(data)
-                input.name = self._input_path1
-
-                if self._input_path2:
-                    data = self._read_pipe.recv_bytes()
-                    input2 = io.BytesIO(data)
-                    input2.name = self._input_path2
-                else:
-                    input2 = None
-                output = io.BytesIO()
-                output.name = self._orig_outfiles.out.name
-
-                if self._orig_outfiles.out2 is not None:
-                    output2 = io.BytesIO()
-                    output2.name = self._orig_outfiles.out2.name
-                else:
-                    output2 = None
-
-                infiles = InputFiles(input, input2, interleaved=self._interleaved_input)
-                outfiles = OutputFiles(out=output, out2=output2, interleaved=self._orig_outfiles.interleaved, force_fasta=self._orig_outfiles.force_fasta)
-                self._pipeline.set_input(infiles)
-                self._pipeline.set_output(outfiles)
+                infiles = self._make_input_files()
+                outfiles = self._make_output_files()
+                self._pipeline.connect_io(infiles, outfiles)
                 (n, bp1, bp2) = self._pipeline.process_reads()
+                self._pipeline.flush()
                 cur_stats = Statistics().collect(n, bp1, bp2, [], self._pipeline._filters)
                 stats += cur_stats
-
-                output.flush()
-                processed_chunk = output.getvalue()
-
-                self._write_pipe.send(chunk_index)
-                self._write_pipe.send(n)  # no. of reads processed in this chunk
-                self._write_pipe.send_bytes(processed_chunk)
-                if self._orig_outfiles.out2 is not None:
-                    output2.flush()
-                    processed_chunk2 = output2.getvalue()
-                    self._write_pipe.send_bytes(processed_chunk2)
+                self._send_outfiles(outfiles, chunk_index, n)
 
             m = self._pipeline._modifiers
             modifier_stats = Statistics().collect(0, 0, 0 if self._pipeline.paired else None, m, [])
@@ -515,6 +532,47 @@ class WorkerProcess(Process):
             self._write_pipe.send(-2)
             self._write_pipe.send((e, traceback.format_exc()))
 
+    def _make_input_files(self):
+        data = self._read_pipe.recv_bytes()
+        input = io.BytesIO(data)
+
+        if self._two_input_files:
+            data = self._read_pipe.recv_bytes()
+            input2 = io.BytesIO(data)
+        else:
+            input2 = None
+        return InputFiles(input, input2, interleaved=self._interleaved_input)
+
+    def _make_output_files(self):
+        """
+        Using self._orig_outfiles as a template, make a new OutputFiles instance
+        that has BytesIO instances for each non-None output file
+        """
+        output_files = copy.copy(self._orig_outfiles)
+        for attr in (
+            "out", "out2", "untrimmed", "untrimmed2", "too_short", "too_short2", "too_long",
+            "too_long2", "info", "rest", "wildcard"
+        ):
+            orig_outfile = getattr(self._orig_outfiles, attr)
+            if orig_outfile is not None:
+                output = io.BytesIO()
+                output.name = orig_outfile.name
+                setattr(output_files, attr, output)
+
+        return output_files
+
+    def _send_outfiles(self, outfiles: OutputFiles, chunk_index: int, n_reads: int):
+        self._write_pipe.send(chunk_index)
+        self._write_pipe.send(n_reads)
+
+        for f in outfiles:
+            if f is None:
+                continue
+            f.flush()
+            assert isinstance(f, io.BytesIO)
+            processed_chunk = f.getvalue()
+            self._write_pipe.send_bytes(processed_chunk)
+
 
 class OrderedChunkWriter:
     """
@@ -527,10 +585,10 @@ class OrderedChunkWriter:
         self._current_index = 0
         self._outfile = outfile
 
-    def write(self, data, chunk_index):
+    def write(self, data, index):
         """
         """
-        self._chunks[chunk_index] = data
+        self._chunks[index] = data
         while self._current_index in self._chunks:
             self._outfile.write(self._chunks[self._current_index])
             del self._chunks[self._current_index]
@@ -544,9 +602,9 @@ class PipelineRunner(ABC):
     """
     A read processing pipeline
     """
-    def __init__(self, pipeline):
+    def __init__(self, pipeline: Pipeline, progress: Progress):
         self._pipeline = pipeline
-        self._progress = Progress()
+        self._progress = progress
 
     @abstractmethod
     def run(self):
@@ -561,7 +619,7 @@ class ParallelPipelineRunner(PipelineRunner):
     """
     Run a Pipeline in parallel
 
-    - When set_input() is called, a reader process is spawned.
+    - When connect_io() is called, a reader process is spawned.
     - When run() is called, as many worker processes as requested are spawned.
     - In the main process, results are written to the output files in the correct
       order, and statistics are aggregated.
@@ -585,63 +643,49 @@ class ParallelPipelineRunner(PipelineRunner):
         pipeline: Pipeline,
         infiles: InputFiles,
         outfiles: OutputFiles,
+        progress: Progress,
         n_workers: int,
-        buffer_size=4*1024**2
+        buffer_size: int = 4 * 1024**2,
     ):
-        super().__init__(pipeline)
-        self._pipes = []  # the workers read from these
-        self._reader_process = None
-        self._outfiles = None
-        self._input_path1 = None
-        self._input_path2 = None
-        self._interleaved_input = None
+        super().__init__(pipeline, progress)
         self._n_workers = n_workers
-        self._need_work_queue = Queue()
+        self._need_work_queue = Queue()  # type: Queue
         self._buffer_size = buffer_size
         self._assign_input(infiles.file1, infiles.file2, infiles.interleaved)
         self._assign_output(outfiles)
 
-    def _assign_input(self, file1, file2=None, interleaved=False):
-        if self._reader_process is not None:
-            raise RuntimeError('Do not call set_input more than once')
-        self._input_path1 = file1 if type(file1) is str else file1.name
-        self._input_path2 = file2 if type(file2) is str or file2 is None else file2.name
+    def _assign_input(
+        self,
+        file1: BinaryIO,
+        file2: Optional[BinaryIO] = None,
+        interleaved: bool = False,
+    ):
+        self._two_input_files = file2 is not None
         self._interleaved_input = interleaved
+        # the workers read from these connections
         connections = [Pipe(duplex=False) for _ in range(self._n_workers)]
-        self._pipes, connw = zip(*connections)
+        self._connections, connw = zip(*connections)
         try:
             fileno = sys.stdin.fileno()
         except io.UnsupportedOperation:
             # This happens during tests: pytest sets sys.stdin to an object
             # that does not have a file descriptor.
             fileno = -1
-        self._reader_process = Process(target=reader_process, args=(file1, file2, connw,
-            self._need_work_queue, self._buffer_size, fileno))
+        self._reader_process = ReaderProcess(file1, file2, connw,
+            self._need_work_queue, self._buffer_size, fileno)
         self._reader_process.daemon = True
         self._reader_process.start()
 
     @staticmethod
-    def can_output_to(outfiles):
-        return (
-            outfiles.out is not None
-            and outfiles.rest is None
-            and outfiles.info is None
-            and outfiles.wildcard is None
-            and outfiles.too_short is None
-            and outfiles.too_short2 is None
-            and outfiles.too_long is None
-            and outfiles.too_long2 is None
-            and outfiles.untrimmed is None
-            and outfiles.untrimmed2 is None
-            and not outfiles.demultiplex
-        )
+    def can_output_to(outfiles: OutputFiles) -> bool:
+        return outfiles.out is not None and not outfiles.demultiplex
 
-    def _assign_output(self, outfiles):
+    def _assign_output(self, outfiles: OutputFiles):
         if not self.can_output_to(outfiles):
             raise ValueError()
         self._outfiles = outfiles
 
-    def _start_workers(self):
+    def _start_workers(self) -> Tuple[List[WorkerProcess], List[Connection]]:
         workers = []
         connections = []
         for index in range(self._n_workers):
@@ -649,9 +693,9 @@ class ParallelPipelineRunner(PipelineRunner):
             connections.append(conn_r)
             worker = WorkerProcess(
                 index, self._pipeline,
-                self._input_path1, self._input_path2,
+                self._two_input_files,
                 self._interleaved_input, self._outfiles,
-                self._pipes[index], conn_w, self._need_work_queue)
+                self._connections[index], conn_w, self._need_work_queue)
             worker.daemon = True
             worker.start()
             workers.append(worker)
@@ -660,7 +704,7 @@ class ParallelPipelineRunner(PipelineRunner):
     def run(self):
         workers, connections = self._start_workers()
         writers = []
-        for outfile in [self._outfiles.out, self._outfiles.out2]:
+        for outfile in self._outfiles:
             if outfile is None:
                 continue
             writers.append(OrderedChunkWriter(outfile))
@@ -678,8 +722,7 @@ class ParallelPipelineRunner(PipelineRunner):
                         # this happens only when there is an exception sending
                         # the statistics)
                         e, tb_str = connection.recv()
-                        # TODO traceback should only be printed in development
-                        logger.debug('%s', tb_str)
+                        logger.error('%s', tb_str)
                         raise e
                     if stats is None:
                         stats = cur_stats
@@ -691,17 +734,16 @@ class ParallelPipelineRunner(PipelineRunner):
                     # An exception has occurred in the worker
                     e, tb_str = connection.recv()
 
-                    # TODO traceback should only be printed in development
                     # We should use the worker's actual traceback object
                     # here, but traceback objects are not picklable.
-                    logger.debug('%s', tb_str)
+                    logger.error('%s', tb_str)
                     raise e
 
                 # No. of reads processed in this chunk
                 chunk_n = connection.recv()
                 if chunk_n == -2:
                     e, tb_str = connection.recv()
-                    logger.debug('%s', tb_str)
+                    logger.error('%s', tb_str)
                     raise e
                 n += chunk_n
                 self._progress.update(n)
@@ -728,10 +770,15 @@ class SerialPipelineRunner(PipelineRunner):
     Run a Pipeline on a single core
     """
 
-    def __init__(self, pipeline: Pipeline, infiles: InputFiles, outfiles: OutputFiles):
-        super().__init__(pipeline)
-        self._pipeline.set_input(infiles)
-        self._pipeline.set_output(outfiles)
+    def __init__(
+        self,
+        pipeline: Pipeline,
+        infiles: InputFiles,
+        outfiles: OutputFiles,
+        progress: Progress,
+    ):
+        super().__init__(pipeline, progress)
+        self._pipeline.connect_io(infiles, outfiles)
 
     def run(self):
         (n, total1_bp, total2_bp) = self._pipeline.process_reads(progress=self._progress)
