@@ -1,12 +1,12 @@
 """
 Routines for printing a report.
 """
-import sys
 from io import StringIO
-from contextlib import contextmanager
 import textwrap
-from .adapters import Where, EndStatistics
-from .modifiers import QualityTrimmer, NextseqQualityTrimmer, AdapterCutter, PairedAdapterCutter
+from typing import Any, Optional, List
+from .adapters import Where, EndStatistics, AdapterStatistics, ADAPTER_TYPE_NAMES
+from .modifiers import (Modifier, PairedModifier, QualityTrimmer, NextseqQualityTrimmer,
+    AdapterCutter, PairedAdapterCutter, ReverseComplementer)
 from .filters import (NoFilter, PairedNoFilter, TooShortReadFilter, TooLongReadFilter,
     PairedDemultiplexer, CombinatorialDemultiplexer, Demultiplexer, NContentFilter, InfoFileWriter,
     WildcardFileWriter, RestFileWriter)
@@ -19,24 +19,33 @@ def safe_divide(numerator, denominator):
         return numerator / denominator
 
 
+def add_if_not_none(a, b):
+    if a is None:
+        return b
+    if b is None:
+        return a
+    return a + b
+
+
 class Statistics:
-    def __init__(self):
+    def __init__(self) -> None:
         """
         """
-        self.paired = None
+        self.paired = None  # type: Optional[bool]
+        self.did_quality_trimming = None  # type: Optional[bool]
         self.too_short = None
         self.too_long = None
         self.too_many_n = None
-        self.did_quality_trimming = None
+        self.reverse_complemented = None  # type: Optional[int]
         self.n = 0
         self.written = 0
         self.total_bp = [0, 0]
         self.written_bp = [0, 0]
         self.with_adapters = [0, 0]
         self.quality_trimmed_bp = [0, 0]
-        self.adapter_stats = [[], []]
+        self.adapter_stats = [[], []]  # type: List[List[AdapterStatistics]]
 
-    def __iadd__(self, other):
+    def __iadd__(self, other: Any):
         self.n += other.n
         self.written += other.written
 
@@ -49,12 +58,8 @@ class Statistics:
         elif self.did_quality_trimming != other.did_quality_trimming:
             raise ValueError('Incompatible Statistics: did_quality_trimming is not equal')
 
-        def add_if_not_none(a, b):
-            if a is None:
-                return b
-            if b is None:
-                return a
-            return a + b
+        self.reverse_complemented = add_if_not_none(
+            self.reverse_complemented, other.reverse_complemented)
         self.too_short = add_if_not_none(self.too_short, other.too_short)
         self.too_long = add_if_not_none(self.too_long, other.too_long)
         self.too_many_n = add_if_not_none(self.too_many_n, other.too_many_n)
@@ -87,45 +92,52 @@ class Statistics:
             self.paired = True
             self.total_bp[1] = total_bp2
 
-        # Collect statistics from writers/filters
-        for w in writers:
-            if isinstance(w, (InfoFileWriter, RestFileWriter, WildcardFileWriter)):
-                pass
-            elif isinstance(w, (NoFilter, PairedNoFilter, PairedDemultiplexer,
-                    CombinatorialDemultiplexer, Demultiplexer)):
-                self.written += w.written
-                self.written_bp[0] += w.written_bp[0]
-                self.written_bp[1] += w.written_bp[1]
-            elif isinstance(w.filter, TooShortReadFilter):
-                self.too_short = w.filtered
-            elif isinstance(w.filter, TooLongReadFilter):
-                self.too_long = w.filtered
-            elif isinstance(w.filter, NContentFilter):
-                self.too_many_n = w.filtered
+        for writer in writers:
+            self._collect_writer(writer)
         assert self.written is not None
-
-        # Collect statistics from modifiers
-        for m in modifiers:
-            if isinstance(m, PairedAdapterCutter):
-                for i in 0, 1:
-                    self.with_adapters[i] += m.with_adapters
-                    self.adapter_stats[i] = list(m.adapter_statistics[i].values())
-                continue
-            if getattr(m, 'paired', False):
-                modifiers_list = [(0, m._modifier1), (1, m._modifier2)]
-            else:
-                modifiers_list = [(0, m)]
-            for i, modifier in modifiers_list:
-                if isinstance(modifier, (QualityTrimmer, NextseqQualityTrimmer)):
-                    self.quality_trimmed_bp[i] = modifier.trimmed_bases
-                    self.did_quality_trimming = True
-                elif isinstance(modifier, AdapterCutter):
-                    self.with_adapters[i] += modifier.with_adapters
-                    self.adapter_stats[i] = list(modifier.adapter_statistics.values())
+        for modifier in modifiers:
+            self._collect_modifier(modifier)
 
         # For chaining
         return self
 
+    def _collect_writer(self, w):
+        if isinstance(w, (InfoFileWriter, RestFileWriter, WildcardFileWriter)):
+            return
+        elif isinstance(w, (NoFilter, PairedNoFilter, PairedDemultiplexer,
+                CombinatorialDemultiplexer, Demultiplexer)):
+            self.written += w.written
+            self.written_bp[0] += w.written_bp[0]
+            self.written_bp[1] += w.written_bp[1]
+        elif isinstance(w.filter, TooShortReadFilter):
+            self.too_short = w.filtered
+        elif isinstance(w.filter, TooLongReadFilter):
+            self.too_long = w.filtered
+        elif isinstance(w.filter, NContentFilter):
+            self.too_many_n = w.filtered
+
+    def _collect_modifier(self, m: Modifier):
+        if isinstance(m, PairedAdapterCutter):
+            for i in 0, 1:
+                self.with_adapters[i] += m.with_adapters
+                self.adapter_stats[i] = list(m.adapter_statistics[i].values())
+            return
+        if isinstance(m, PairedModifier):
+            modifiers_list = [(0, m._modifier1), (1, m._modifier2)]
+        else:
+            modifiers_list = [(0, m)]
+        for i, modifier in modifiers_list:
+            if isinstance(modifier, (QualityTrimmer, NextseqQualityTrimmer)):
+                self.quality_trimmed_bp[i] = modifier.trimmed_bases
+                self.did_quality_trimming = True
+            elif isinstance(modifier, AdapterCutter):
+                self.with_adapters[i] += modifier.with_adapters
+                self.adapter_stats[i] = list(modifier.adapter_statistics.values())
+            elif isinstance(modifier, ReverseComplementer):
+                self.with_adapters[i] += modifier.adapter_cutter.with_adapters
+                self.adapter_stats[i] = list(modifier.adapter_cutter.adapter_statistics.values())
+                self.reverse_complemented = modifier.reverse_complemented
+
     @property
     def total(self):
         return sum(self.total_bp)
@@ -154,6 +166,10 @@ class Statistics:
     def total_written_bp_fraction(self):
         return safe_divide(self.total_written_bp, self.total)
 
+    @property
+    def reverse_complemented_fraction(self):
+        return safe_divide(self.reverse_complemented, self.n)
+
     @property
     def too_short_fraction(self):
         return safe_divide(self.too_short, self.n)
@@ -167,18 +183,6 @@ class Statistics:
         return safe_divide(self.too_many_n, self.n)
 
 
-ADAPTER_TYPES = {
-    Where.BACK: "regular 3'",
-    Where.BACK_NOT_INTERNAL: "non-internal 3'",
-    Where.FRONT: "regular 5'",
-    Where.FRONT_NOT_INTERNAL: "non-internal 5'",
-    Where.PREFIX: "anchored 5'",
-    Where.SUFFIX: "anchored 3'",
-    Where.ANYWHERE: "variable 5'/3'",
-    Where.LINKED: "linked",
-}
-
-
 def error_ranges(adapter_statistics: EndStatistics):
     length = adapter_statistics.effective_length
     error_rate = adapter_statistics.max_error_rate
@@ -271,7 +275,7 @@ class AdjacentBaseStatistics:
         return sio.getvalue()
 
 
-def full_report(stats, time, gc_content) -> str:
+def full_report(stats: Statistics, time: float, gc_content: float) -> str:  # noqa: C901
     """Print report to standard output."""
     if stats.n == 0:
         return "No reads processed!"
@@ -297,6 +301,9 @@ def full_report(stats, time, gc_content) -> str:
         Total reads processed:           {o.n:13,d}
         Reads with adapters:             {o.with_adapters[0]:13,d} ({o.with_adapters_fraction[0]:.1%})
         """)
+    if stats.reverse_complemented is not None:
+        report += "Reverse-complemented:            " \
+                  "{o.reverse_complemented:13,d} ({o.reverse_complemented_fraction:.1%})\n"
     if stats.too_short is not None:
         report += "{pairs_or_reads} that were too short:       {o.too_short:13,d} ({o.too_short_fraction:.1%})\n"
     if stats.too_long is not None:
@@ -333,10 +340,13 @@ def full_report(stats, time, gc_content) -> str:
             total_front = sum(adapter_statistics.front.lengths.values())
             total_back = sum(adapter_statistics.back.lengths.values())
             total = total_front + total_back
+            reverse_complemented = adapter_statistics.reverse_complemented
             where = adapter_statistics.where
+            where_backs = (Where.BACK, Where.BACK_NOT_INTERNAL, Where.SUFFIX)
+            where_fronts = (Where.FRONT, Where.FRONT_NOT_INTERNAL, Where.PREFIX)
             assert (where in (Where.ANYWHERE, Where.LINKED)
-                or (where in (Where.BACK, Where.BACK_NOT_INTERNAL, Where.SUFFIX) and total_front == 0)
-                or (where in (Where.FRONT, Where.FRONT_NOT_INTERNAL, Where.PREFIX) and total_back == 0)), (where, total_front)
+                or (where in where_backs and total_front == 0)
+                or (where in where_fronts and total_back == 0)), (where, total_front, total_back)
 
             if stats.paired:
                 extra = 'First read: ' if which_in_pair == 0 else 'Second read: '
@@ -355,9 +365,13 @@ def full_report(stats, time, gc_content) -> str:
                         len(adapter_statistics.back.sequence),
                         total_front, total_back))
             else:
-                print_s("Sequence: {}; Type: {}; Length: {}; Trimmed: {} times.".
-                    format(adapter_statistics.front.sequence, ADAPTER_TYPES[adapter_statistics.where],
-                        len(adapter_statistics.front.sequence), total))
+                print_s("Sequence: {}; Type: {}; Length: {}; Trimmed: {} times".
+                    format(adapter_statistics.front.sequence, ADAPTER_TYPE_NAMES[adapter_statistics.where],
+                        len(adapter_statistics.front.sequence), total), end="")
+            if reverse_complemented is not None:
+                print_s("; Reverse-complemented: {} times".format(reverse_complemented))
+            else:
+                print_s()
             if total == 0:
                 print_s()
                 continue
@@ -403,7 +417,7 @@ def full_report(stats, time, gc_content) -> str:
     return sio.getvalue().rstrip()
 
 
-def minimal_report(stats, time, gc_content) -> str:
+def minimal_report(stats, _time, _gc_content) -> str:
     """Create a minimal tabular report suitable for concatenation"""
 
     def none(value):
