@@ -1,8 +1,12 @@
 import re
 import sys
 import time
+import errno
 import multiprocessing
 
+from xopen import xopen
+import dnaio
+
 
 def available_cpu_count():
     """
@@ -27,22 +31,26 @@ def available_cpu_count():
     return multiprocessing.cpu_count()
 
 
+def raise_open_files_limit(n):
+    try:
+        import resource
+    except ImportError:
+        # Windows
+        return
+    soft, hard = resource.getrlimit(resource.RLIMIT_NOFILE)
+    soft += n
+    resource.setrlimit(resource.RLIMIT_NOFILE, (soft, hard))
+
+
 class Progress:
     """
-    Write progress
+    Print an animated progress report to sys.stderr
     """
 
     def __init__(self, every=1):
         """
-        :param every: minimum time to wait in seconds between progress updates
+        every: minimum time to wait in seconds between progress updates
         """
-        if not sys.stderr.isatty():
-            # Disable progress
-            self.update = self.do_nothing
-            self._show = False
-        else:
-            self._show = True
-
         self._every = every
         self._n = 0
         # Time at which the progress was last updated
@@ -65,12 +73,7 @@ class Progress:
                             sc = "8<" if is_open else "8="
                         yield "[" + left + sc + right + "]"
 
-    def do_nothing(self, total, _final=False):
-        pass
-
     def update(self, total, _final=False):
-        """
-        """
         current_time = time.time()
         if _final:
             time_delta = current_time - self._start_time
@@ -111,5 +114,78 @@ class Progress:
         Print final progress reflecting the final total
         """
         self.update(total, _final=True)
-        if self._show:
-            print(file=sys.stderr)
+        print(file=sys.stderr)
+
+
+class DummyProgress:
+    """
+    Has the same interface as Progress, but does not print anything
+    """
+    def update(self, total, _final=False):
+        pass
+
+    def stop(self, total):
+        pass
+
+
+_REVCOMP_TRANS = str.maketrans(
+    "ACGTUMRWSYKVHDBNacgtumrwsykvhdbn",
+    "TGCAAKYWSRMBDHVNtgcaakywsrmbdhvn",
+)
+
+
+def reverse_complement(s: str):
+    return s.translate(_REVCOMP_TRANS)[::-1]
+
+
+def reverse_complemented_sequence(sequence: dnaio.Sequence):
+    if sequence.qualities is None:
+        qualities = None
+    else:
+        qualities = sequence.qualities[::-1]
+    return dnaio.Sequence(sequence.name, reverse_complement(sequence.sequence), qualities)
+
+
+class FileOpener:
+    def __init__(self, compression_level: int = 6, threads: int = None):
+        self.compression_level = compression_level
+        self.threads = threads
+
+    def xopen(self, path, mode):
+        return xopen(path, mode, compresslevel=self.compression_level, threads=self.threads)
+
+    def xopen_or_none(self, path, mode):
+        """Return opened file or None if the path is None"""
+        if path is None:
+            return None
+        return self.xopen(path, mode)
+
+    def xopen_pair(self, path1, path2, mode):
+        file1 = file2 = None
+        if path1 is not None:
+            file1 = self.xopen(path1, mode)
+            if path2 is not None:
+                file2 = self.xopen(path2, mode)
+        elif path2 is not None:
+            raise ValueError("When giving paths for paired-end files, only providing the second"
+                " file is not supported")
+        return file1, file2
+
+    def dnaio_open(self, *args, **kwargs):
+        kwargs["opener"] = self.xopen
+        return dnaio.open(*args, **kwargs)
+
+    def dnaio_open_raise_limit(self, path, qualities):
+        """
+        Open a FASTA/FASTQ file for writing. If it fails because the number of open files
+        would be exceeded, try to raise the soft limit and re-try.
+        """
+        try:
+            f = self.dnaio_open(path, mode="w", qualities=qualities)
+        except OSError as e:
+            if e.errno == errno.EMFILE:  # Too many open files
+                raise_open_files_limit(8)
+                f = self.dnaio_open(path, mode="w", qualities=qualities)
+            else:
+                raise
+        return f
