@@ -9,6 +9,8 @@ from cutadapt.align import (
 )
 from cutadapt.adapters import Where
 
+from utils import binomial
+
 
 # convenience function (to avoid having to instantiate an Aligner manually)
 def locate(reference, query, max_error_rate, flags=SEMIGLOBAL, wildcard_ref=False,
@@ -196,7 +198,7 @@ def test_wildcards_in_adapter():
 
 def test_wildcards_in_read():
     a = WILDCARD_SEQUENCES[0]
-    for s in WILDCARD_SEQUENCES:
+    for s in WILDCARD_SEQUENCES + ['CCCXTTXATC']:
         r = 'CATCTGTCC' + s + 'GCCAGGGTTGATTCGGCTGATCTGGCCG'
         result = locate(a, r, 0.0, Where.BACK.value, wildcard_query=True)
         if 'X' in s:
@@ -208,8 +210,6 @@ def test_wildcards_in_read():
 def test_wildcards_in_both():
     for a in WILDCARD_SEQUENCES:
         for s in WILDCARD_SEQUENCES:
-            if 'X' in s or 'X' in a:
-                continue
             r = 'CATCTGTCC' + s + 'GCCAGGGTTGATTCGGCTGATCTGGCCG'
             result = locate(a, r, 0.0, Where.BACK.value, wildcard_ref=True, wildcard_query=True)
             assert result == (0, 10, 9, 19, 10, 0), result
@@ -220,21 +220,6 @@ def test_no_match():
     assert a is None, a
 
 
-def binomial(n, k):
-    """
-    Return binomial coefficient ('n choose k').
-    This implementation does not use factorials.
-    """
-    k = min(k, n - k)
-    if k < 0:
-        return 0
-    r = 1
-    for j in range(k):
-        r *= n - j
-        r //= j + 1
-    return r
-
-
 def test_hamming_sphere_explicit():
     assert list(hamming_sphere('', 0)) == ['']
     assert list(hamming_sphere('A', 0)) == ['A']
