@@ -2,7 +2,6 @@
 # test reading from standard input
 import os
 import shutil
-import subprocess
 import sys
 import tempfile
 from io import StringIO
@@ -15,7 +14,7 @@ from utils import assert_files_equal, datapath, cutpath, redirect_stderr
 # pytest.mark.timeout will not fail even if pytest-timeout is not installed
 try:
     import pytest_timeout as _unused
-except ImportError:
+except ImportError:  # pragma: no cover
     raise ImportError("pytest_timeout needs to be installed")
 del _unused
 
@@ -32,9 +31,9 @@ def test_small(run):
     run('-a TTAGACATATCTCCGTCG', 'small.fastq', 'small.fastq')
 
 
-def test_empty(run):
+def test_empty(run, cores):
     """empty input"""
-    run('-a TTAGACATATCTCCGTCG', 'empty.fastq', 'empty.fastq')
+    run("--cores {} -a TTAGACATATCTCCGTCG".format(cores), "empty.fastq", "empty.fastq")
 
 
 def test_newlines(run):
@@ -47,10 +46,10 @@ def test_lowercase(run):
     run('-a ttagacatatctccgtcg', 'lowercase.fastq', 'small.fastq')
 
 
-def test_rest(run, tmpdir):
+def test_rest(run, tmpdir, cores):
     """-r/--rest-file"""
     rest = str(tmpdir.join("rest.tmp"))
-    run(['-b', 'ADAPTER', '-N', '-r', rest], "rest.fa", "rest.fa")
+    run(['--cores', str(cores), '-b', 'ADAPTER', '-N', '-r', rest], "rest.fa", "rest.fa")
     assert_files_equal(datapath('rest.txt'), rest)
 
 
@@ -98,10 +97,15 @@ def test_minimum_length(run):
     run("-m 5 -a TTAGACATATCTCCGTCG", "minlen.fa", "lengths.fa")
 
 
-def test_too_short(run, tmpdir):
+def test_too_short(run, tmpdir, cores):
     """--too-short-output"""
     too_short_path = str(tmpdir.join('tooshort.fa'))
-    run("-m 5 -a TTAGACATATCTCCGTCG --too-short-output " + too_short_path, "minlen.fa", "lengths.fa")
+    run([
+        "--cores", str(cores),
+        "-m", "5",
+        "-a", "TTAGACATATCTCCGTCG",
+        "--too-short-output", too_short_path
+    ], "minlen.fa", "lengths.fa")
     assert_files_equal(datapath('tooshort.fa'), too_short_path)
 
 
@@ -110,10 +114,15 @@ def test_maximum_length(run):
     run("-M 5 -a TTAGACATATCTCCGTCG", "maxlen.fa", "lengths.fa")
 
 
-def test_too_long(run, tmpdir):
+def test_too_long(run, tmpdir, cores):
     """--too-long-output"""
     too_long_path = str(tmpdir.join('toolong.fa'))
-    run("-M 5 -a TTAGACATATCTCCGTCG --too-long-output " + too_long_path, "maxlen.fa", "lengths.fa")
+    run([
+        "--cores", str(cores),
+        "-M", "5",
+        "-a", "TTAGACATATCTCCGTCG",
+        "--too-long-output", too_long_path
+    ], "maxlen.fa", "lengths.fa")
     assert_files_equal(datapath('toolong.fa'), too_long_path)
 
 
@@ -219,11 +228,14 @@ def test_read_wildcard(run):
     ("-a", "wildcard_adapter.fa"),
     ("-b", "wildcard_adapter_anywhere.fa"),
 ])
-def test_adapter_wildcard(adapter_type, expected, run, tmpdir):
+def test_adapter_wildcard(adapter_type, expected, run, tmpdir, cores):
     """wildcards in adapter"""
     wildcard_path = str(tmpdir.join("wildcards.txt"))
-    run("--wildcard-file {} {} ACGTNNNACGT".format(wildcard_path, adapter_type),
-        expected, "wildcard_adapter.fa")
+    run([
+            "--cores", str(cores),
+            "--wildcard-file", wildcard_path,
+            adapter_type, "ACGTNNNACGT"
+        ],  expected, "wildcard_adapter.fa")
     with open(wildcard_path) as wct:
         lines = wct.readlines()
     lines = [line.strip() for line in lines]
@@ -300,26 +312,26 @@ def test_strip_suffix(run):
     run("--strip-suffix _sequence -a XXXXXXX", "stripped.fasta", "simple.fasta")
 
 
-def test_info_file(run, tmpdir):
+def test_info_file(run, tmpdir, cores):
     # The true adapter sequence in the illumina.fastq.gz data set is
     # GCCTAACTTCTTAGACTGCCTTAAGGACGT (fourth base is different from the sequence shown here)
     info_path = str(tmpdir.join("info.txt"))
-    run(["--info-file", info_path, "-a", "adapt=GCCGAACTTCTTAGACTGCCTTAAGGACGT"],
+    run(["--cores", str(cores), "--info-file", info_path, "-a", "adapt=GCCGAACTTCTTAGACTGCCTTAAGGACGT"],
         "illumina.fastq", "illumina.fastq.gz")
     assert_files_equal(cutpath("illumina.info.txt"), info_path)
 
 
-def test_info_file_times(run, tmpdir):
+def test_info_file_times(run, tmpdir, cores):
     info_path = str(tmpdir.join("info.txt"))
-    run(["--info-file", info_path, "--times", "2", "-a", "adapt=GCCGAACTTCTTA",
+    run(["--cores", str(cores), "--info-file", info_path, "--times", "2", "-a", "adapt=GCCGAACTTCTTA",
         "-a", "adapt2=GACTGCCTTAAGGACGT"], "illumina5.fastq", "illumina5.fastq")
     assert_files_equal(cutpath('illumina5.info.txt'), info_path)
 
 
-def test_info_file_fasta(run, tmpdir):
+def test_info_file_fasta(run, tmpdir, cores):
     info_path = str(tmpdir.join("info.txt"))
     # Just make sure that it runs
-    run(["--info-file", info_path, "-a", "TTAGACATAT", "-g", "GAGATTGCCA", "--no-indels"],
+    run(["--cores", str(cores), "--info-file", info_path, "-a", "TTAGACATAT", "-g", "GAGATTGCCA", "--no-indels"],
         "no_indels.fasta", "no_indels.fasta")
 
 
@@ -340,13 +352,8 @@ if sys.version_info[:2] >= (3, 3):
         run('-b TTAGACATATCTCCGTCG', 'small.fastq', 'multiblock.fastq.bz2')
 
 
-try:
-    import lzma
-
-    def test_xz(run):
-        run('-b TTAGACATATCTCCGTCG', 'small.fastq', 'small.fastq.xz')
-except ImportError:
-    pass
+def test_xz(run):
+    run('-b TTAGACATATCTCCGTCG', 'small.fastq', 'small.fastq.xz')
 
 
 def test_no_args():
@@ -368,7 +375,8 @@ def test_anchored_no_indels(run):
 
 def test_anchored_no_indels_wildcard_read(run):
     """anchored 5' adapter, mismatches only (no indels), but wildcards in the read count as matches"""
-    run('-g ^TTAGACATAT --match-read-wildcards --no-indels -e 0.1', 'anchored_no_indels_wildcard.fasta', 'anchored_no_indels.fasta')
+    run('-g ^TTAGACATAT --match-read-wildcards --no-indels -e 0.1',
+        'anchored_no_indels_wildcard.fasta', 'anchored_no_indels.fasta')
 
 
 def test_anchored_no_indels_wildcard_adapt(run):
@@ -394,9 +402,20 @@ def test_unconditional_cut_both(run):
     run('-u -5 -u 5', 'unconditional-both.fastq', 'small.fastq')
 
 
-def test_untrimmed_output(run, tmpdir):
+def test_unconditional_cut_too_many_commas():
+    with pytest.raises(SystemExit):
+        main(["-u", "5,7,8", datapath("small.fastq")])
+
+
+def test_unconditional_cut_invalid_number():
+    with pytest.raises(SystemExit):
+        main(["-u", "a,b", datapath("small.fastq")])
+
+
+def test_untrimmed_output(run, cores, tmpdir):
     path = str(tmpdir.join("untrimmed.fastq"))
-    run(["-a", "TTAGACATATCTCCGTCG", "--untrimmed-output", path], "small.trimmed.fastq", "small.fastq")
+    run(["--cores", str(cores), "-a", "TTAGACATATCTCCGTCG", "--untrimmed-output", path],
+        "small.trimmed.fastq", "small.fastq")
     assert_files_equal(cutpath("small.untrimmed.fastq"), path)
 
 
@@ -511,6 +530,18 @@ def test_linked_discard_untrimmed_g(run):
     run('-g AAAAAAAAAA...TTTTTTTTTT --discard-untrimmed', 'linked-discard-g.fasta', 'linked.fasta')
 
 
+def test_linked_lowercase(run):
+    run('-a ^AACCGGTTTT...GGGGGGG$ -a ^AAAA...TTTT$ --times=2 --action=lowercase',
+        'linked-lowercase.fasta', 'linked.fasta')
+
+
+def test_linked_info_file(tmpdir):
+    info_path = str(tmpdir.join('info.txt'))
+    main(['-a linkedadapter=^AAAAAAAAAA...TTTTTTTTTT', '--info-file', info_path,
+        '-o', str(tmpdir.join('out.fasta')), datapath('linked.fasta')])
+    assert_files_equal(cutpath('linked-info.txt'), info_path)
+
+
 def test_linked_anywhere():
     with pytest.raises(SystemExit):
         with redirect_stderr():
@@ -666,3 +697,27 @@ def test_force_fasta_output(tmpdir, cores):
 
 def test_empty_read_with_wildcard_in_adapter(run):
     run("-g CWC", "empty.fastq", "empty.fastq")
+
+
+def test_print_progress_to_tty(tmpdir, mocker):
+    mocker.patch("cutadapt.utils.sys.stderr").isatty.return_value = True
+    main(["-o", str(tmpdir.join("out.fastq")), datapath("small.fastq")])
+
+
+def test_adapter_order(run):
+    run("-g ^AAACC -a CCGGG", "adapterorder-ga.fasta", "adapterorder.fasta")
+    run("-a CCGGG -g ^AAACC", "adapterorder-ag.fasta", "adapterorder.fasta")
+
+
+@pytest.mark.skip(reason="Not implemented")
+def test_reverse_complement_not_normalized(run):
+    run("--rc=yes -g ^TTATTTGTCT -g ^TCCGCACTGG",
+        "revcomp-notnormalized-single.fastq", "revcomp.1.fastq")
+
+
+def test_reverse_complement_normalized(run):
+    run(
+        "--revcomp -g ^TTATTTGTCT -g ^TCCGCACTGG",
+        "revcomp-single-normalize.fastq",
+        "revcomp.1.fastq",
+    )
