@@ -30,3 +30,18 @@ def assert_files_equal(path1, path2):
         subprocess.check_output(['diff', '-u', path1, path2], stderr=subprocess.STDOUT)
     except subprocess.CalledProcessError as e:
         raise FilesDifferent('\n' + e.output.decode()) from None
+
+
+def binomial(n, k):
+    """
+    Return binomial coefficient ('n choose k').
+    This implementation does not use factorials.
+    """
+    k = min(k, n - k)
+    if k < 0:
+        return 0
+    r = 1
+    for j in range(k):
+        r *= n - j
+        r //= j + 1
+    return r
